(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2009-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: options.ml,v 1.57 2017/05/31 16:38:15 fclement Exp $ *)

module Nat = Unatural;;
module Cmd_opt = Ucommand_option;;

let _0_ = Nat._0_;;

open Uoption_key;;

let string_of_list s_of l =
  let len = Ulist.length l in
  Ulist.fold_left_i
    (fun i accu x ->
      if i = _0_ then accu ^ s_of x else
      if i = Nat.pred len then accu ^ ", and " ^ s_of x else
      accu ^ ", " ^ s_of x)
    "" l
;;

let o_qual = Upervasives.qualify "Options";;

let o_failwith = Upervasives.failwith_qual o_qual;;
let o_invalid_arg = Upervasives.invalid_arg_qual o_qual;;

open Segmentation_types;;

(* Codomain implementation.*)

type codomain_implementation =
  | Tuple_Cod
  | Vector_Cod
;;

let default_codomain_implementation = Tuple_Cod;;

let codomain_implementation_sok_st = Short (Uoption_key.of_string "-impl");;

module Codomain_implementation_structure =
struct
  type t = codomain_implementation;;
  
  let name = "codomain implementation"
  and plural_name = "codomain implementations"
  ;;
  
  let default = default_codomain_implementation;;
  
  let of_string = function
    | "tuple_codomain" | "tuple" | "tc" | "t" -> Tuple_Cod
    | "vector_codomain" | "vector" | "vc" | "v" -> Vector_Cod
    | s ->
        o_failwith "codomain_implementation_of_string"
          (Printf.sprintf "Unknown %s %S" name s)
  ;;
  
  let to_string = function
    | Tuple_Cod -> "tuple_codomain"
    | Vector_Cod -> "vector_codomain"
  ;;
  
  let all = [ Tuple_Cod; Vector_Cod; ];;
end
;;

module Codomain_implementation =
  Ustring_command_option.Make (Codomain_implementation_structure)
;;
let get_codomain_implementation = Codomain_implementation.get
and codomain_implementation_lok = Codomain_implementation.lok
and codomain_implementation_rco = Codomain_implementation.rco
;;

(* Dimension.*)

type dimension = int;;

let dimension_option_name = "dimension"
and default_dimension = 3
;;

let { Uaref.get = get_dimension;
      set = set_dimension; } =
  Uaref.make_aref_single dimension_option_name default_dimension
;;

let { long = dimension_lok; short = dimension_sok_st; } =
  make_option_keys dimension_option_name (Short (of_string "-dim"))
;;

let dimension_rco =
  {
   Cmd_opt.key = dimension_lok;
   spec = Arg.Int set_dimension;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet dimension of the image in each pixel to <nat>.\n\
        \tThe dimension is a positive natural number.\n\
        \tIt is 1 for monochrome images, and 3 for RGB images.\n\
        \tDefault dimension is %d."
       (get_dimension ());
  }
;;

(* Dimension management. *)

type dimension_management =
  | Vector
  | Best_component_only
  | Best_component_for_each
  | Combine_best_components
;;

let default_dimension_management = Vector;;

let dimension_management_sok_st = Automatic;;

module Dimension_management_structure =
struct
  type t = dimension_management;;
  
  let name = "dimension management"
  and plural_name = "dimension managements"
  ;;
  
  let default = default_dimension_management;;
  
  let of_string = function
    | "vector" | "v" ->
      Vector
    | "best_component_only" | "best_only" | "bco" ->
      Best_component_only
    | "best_component_for_each" | "each" | "bcfe" ->
      Best_component_for_each
    | "combine_best_components" | "combine" | "cbc" ->
      Combine_best_components
    | s ->
      o_failwith "dimension_management_of_string"
        (Printf.sprintf "Unknown %s %S" name s)
  ;;
  
  let to_string = function
    | Vector -> "vector"
    | Best_component_only -> "best_component_only"
    | Best_component_for_each -> "best_component_for_each"
    | Combine_best_components -> "combine_best_components"
  ;;
  
  let all =
    [ Vector; Best_component_only;
      Best_component_for_each; Combine_best_components; ]
  ;;
end
;;

module Dimension_management =
  Ustring_command_option.Make (Dimension_management_structure)
;;
let get_dimension_management = Dimension_management.get
and dimension_management_rco = Dimension_management.rco
;;

(* Cutting strategy. *)

type cutting_strategy =
  | Option_Best_in_a_family
  | Option_Overall_best
;;

let default_cutting_strategy = Option_Overall_best;;

let cutting_strategy_sok_st = Automatic;;

module Cutting_strategy_structure =
struct
  type t = cutting_strategy;;
  
  let name = "cutting strategy"
  and plural_name = "cutting strategies"
  ;;
  
  let default = default_cutting_strategy;;
  
  let of_string = function
    | "best_in_a_family" | "bf" | "family" | "f" -> Option_Best_in_a_family
    | "overall_best" | "ob" | "best" | "b" -> Option_Overall_best
    | s ->
        o_failwith "cutting_strategy_of_string"
          (Printf.sprintf "Unknown %s %S" name s)
  ;;
  
  let to_string = function
    | Option_Best_in_a_family -> "best_in_a_family"
    | Option_Overall_best -> "overall_best"
  ;;
  
  let all = [ Option_Best_in_a_family; Option_Overall_best; ];;
end
;;

module type CUTTING_STRATEGY =
    Ustring_command_option_spec.T with type t = cutting_strategy
;;
module Cutting_strategy : CUTTING_STRATEGY =
  Ustring_command_option.Make (Cutting_strategy_structure)
;;
let get_cutting_strategy = Cutting_strategy.get
and cutting_strategy_rco = Cutting_strategy.rco
;;

type number_of_cuts = int;;

let number_of_cuts_option_name = "number of cuts"
and default_number_of_cuts = 0
;;

let { Uaref.get = get_number_of_cuts;
      set = set_number_of_cuts; } =
  Uaref.make_aref_single number_of_cuts_option_name default_number_of_cuts
;;

let { long = number_of_cuts_lok; short = number_of_cuts_sok_st; } =
  make_option_keys number_of_cuts_option_name (Short (of_string "-nc"))
;;

let number_of_cuts_rco =
  {
   Cmd_opt.key = number_of_cuts_lok;
   spec = Arg.Int set_number_of_cuts;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet number of cuts for each basic cutting family to <nat>.\n\
        \tThe number of cuts is a natural number.\n\
        \tBy convention, 0 means all possible cuts of the family.\n\
        \tDefault number of cuts is %d."
       (get_number_of_cuts ());
  }
;;

type basic_cutting_families = basic_cutting_family list;;

let all_basic_cutting_families = [ Horizontal; Vertical; ];;

let clear_basic_cutting_families_option_name = "clear basic cutting families"
and default_basic_cutting_families = all_basic_cutting_families
;;

let { Uaref.get_list = get_basic_cutting_families;
      clear_list = clear_basic_cutting_families;
      add_item = add_basic_cutting_family; } =
  Uaref.make_aref_list default_basic_cutting_families
;;

let { long = clear_basic_cutting_families_lok;
      short = clear_basic_cutting_families_sok_st; } =
  make_option_keys
    clear_basic_cutting_families_option_name
    (Short (of_string "-clear-bcf"))
;;

let clear_basic_cutting_families_rco =
  {
   Cmd_opt.key = clear_basic_cutting_families_lok;
   spec = Arg.Unit clear_basic_cutting_families;
   doc = Printf.sprintf "\n\tClear the basic cutting family list.";
  }
;;

let add_basic_cutting_family_option_name = "add basic cutting family";;

let basic_cutting_family_of_string = function
  | "horizontal" | "hor" | "h" -> Horizontal
  | "vertical" | "ver" | "v" -> Vertical
  | s ->
      o_failwith "basic_cutting_family_of_string"
        (Printf.sprintf "Unknown basic cutting family %S" s)
;;

let add_basic_cutting_family_from_string =
  Upervasives.compose
    add_basic_cutting_family
    basic_cutting_family_of_string
;;

let string_of_basic_cutting_family = function
  | Horizontal -> "horizontal"
  | Vertical -> "vertical"
;;

let print_all_basic_cutting_families () =
  string_of_list
    string_of_basic_cutting_family
    all_basic_cutting_families
;;

let print_default_basic_cutting_family_list () =
  string_of_list
    string_of_basic_cutting_family
    (get_basic_cutting_families ())
;;

let { long = add_basic_cutting_family_lok;
      short = add_basic_cutting_family_sok_st; } =
  make_option_keys
    add_basic_cutting_family_option_name (Short (of_string "-add-bcf"))
;;

let add_basic_cutting_family_rco =
  {
   Cmd_opt.key = add_basic_cutting_family_lok;
   spec = Arg.String add_basic_cutting_family_from_string;
   doc =
     Printf.sprintf
       "<str>\n\
        \tAdd <str> to the list of basic cutting families.\n\
        \tUsed when the cutting strategy is set to %S.\n\
        \tValid basic cutting families are: %t.\n\
        \tDefault list is %t."
       (Cutting_strategy_structure.to_string Option_Best_in_a_family)
       print_all_basic_cutting_families
       print_default_basic_cutting_family_list;
  }
;;

(* Stopping criterion. *)

type max_iteration = int;;

let max_iteration_option_name = "max iteration"
and default_max_iteration = 10
;;

let { Uaref.get = get_max_iteration;
      set = set_max_iteration; } =
  Uaref.make_aref_single max_iteration_option_name default_max_iteration
;;

let { long = max_iteration_lok; short = max_iteration_sok_st; } =
  make_option_keys max_iteration_option_name (Short (of_string "-max-it"))
;;

let max_iteration_rco =
  {
   Cmd_opt.key = max_iteration_lok;
   spec = Arg.Int set_max_iteration;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet maximum number of iterations to <nat>.\n\
        \tIteration numbers are natural numbers.\n\
        \tDefault maximum number of iterations is %d."
       (get_max_iteration ());
  }
;;

type min_best_scalar_indicator = float;;

let min_best_scalar_indicator_option_name = "min best scalar indicator"
and default_min_best_scalar_indicator = 1e-2
;;

let { Uaref.get = get_min_best_scalar_indicator;
      set = set_min_best_scalar_indicator; } =
  Uaref.make_aref_single
    min_best_scalar_indicator_option_name
    default_min_best_scalar_indicator
;;

let { long = min_best_scalar_indicator_lok;
      short = min_best_scalar_indicator_sok_st; } =
  make_option_keys
    min_best_scalar_indicator_option_name (Short (of_string "-min-ind"))
;;

let min_best_scalar_indicator_rco =
  {
   Cmd_opt.key = min_best_scalar_indicator_lok;
   spec = Arg.Float set_min_best_scalar_indicator;
   doc =
     Printf.sprintf
       "<float>\n\
        \tSet minimum best scalar indicator to <float>.\n\
        \tIndicators are non negative real numbers.\n\
        \tDefault minimum best scalar indicator is %g."
       (get_min_best_scalar_indicator ());
  }
;;

type min_scalar_relative_error = float;;

let min_scalar_relative_error_option_name = "min scalar relative error"
and default_min_scalar_relative_error = 1e-2
;;

let { Uaref.get = get_min_scalar_relative_error;
      set = set_min_scalar_relative_error; } =
  Uaref.make_aref_single
    min_scalar_relative_error_option_name
    default_min_scalar_relative_error
;;

let { long = min_scalar_relative_error_lok;
      short = min_scalar_relative_error_sok_st; } =
  make_option_keys
    min_scalar_relative_error_option_name (Short (of_string "-min-err"))
;;

let min_scalar_relative_error_rco =
  {
   Cmd_opt.key = min_scalar_relative_error_lok;
   spec = Arg.Float set_min_scalar_relative_error;
   doc =
     Printf.sprintf
       "<float>\n\
        \tSet minimum scalar relative error to <float>.\n\
        \tRelative errors are non negative real numbers.\n\
        \tDefault minimum scalar relative error is %g."
       (get_min_scalar_relative_error ());
  }
;;

(* Program configuration. *)

type out_formatter =
  | Std_out
  | Std_err
  | Out_file of Ufilename.t
;;

let default_out_formatter = Std_out;;

let out_formatter_sok_st = Short (Uoption_key.of_string "-log");;

module Out_formatter_structure =
struct
  type t = out_formatter;;
  
  let name = "out formatter"
  and plural_name = "out formatters"
  ;;
  
  let default = default_out_formatter;;
  
  let of_string = function
    | "-" | "stdout" | "Std_out" -> Std_out
    | "stderr" | "Std_err" -> Std_err
    | "" ->
        o_invalid_arg "out_formatter_of_string"
          "Empty out formatter specification"
    | s -> Out_file s
  ;;
  
  let to_string = function
    | Std_out -> "Std_out"
    | Std_err -> "Std_err"
    | Out_file fname -> fname
  ;;
  
  let all = [ Std_out; Std_err; Out_file "<file name>" ; ];;
end
;;

module type OUT_FORMATTER =
    Ustring_command_option_spec.T with type t = out_formatter
;;
module Out_formatter : OUT_FORMATTER =
  Ustring_command_option.Make (Out_formatter_structure)
;;
let get_out_formatter = Out_formatter.get
and out_formatter_rco = Out_formatter.rco
;;

type report_allocated_memory = bool;;

let report_allocated_memory_option_name = "report allocated memory";;

let { Uaref.get_flag = get_report_allocated_memory;
      set_flag = set_report_allocated_memory; _; } =
  Uaref.make_flag ()
;;

let { long = report_allocated_memory_lok;
      short = report_allocated_memory_sok_st; } =
  make_option_keys report_allocated_memory_option_name Automatic
;;

let report_allocated_memory_rco =
  {
   Cmd_opt.key = report_allocated_memory_lok;
   spec = Arg.Unit set_report_allocated_memory;
   doc =
     Printf.sprintf
       "\n\tReport total allocated memory. Default is no report.";
  }
;;

type incremental_mode = bool;;

let incremental_mode_option_name = "incremental mode";;

let { Uaref.get_flag = get_incremental_mode;
      set_flag = set_incremental_mode; _; } =
  Uaref.make_flag ()
;;

let { long = incremental_mode_lok; short = incremental_mode_sok_st; } =
  make_option_keys incremental_mode_option_name Automatic
;;

let incremental_mode_rco =
  {
   Cmd_opt.key = incremental_mode_lok;
   spec = Arg.Unit set_incremental_mode;
   doc =
     Printf.sprintf "\n\tTurn the incremental mode on. Default is off.";
  }
;;

type graphics_mode =
  | No_graphics
  | Minimal_graphics
  | Parameter_graphics
  | Compound_graphics
  | Full_graphics
;;

let default_graphics_mode = Full_graphics;;

let graphics_mode_sok_st = Automatic;;

module Graphics_mode_structure =
struct
  type t = graphics_mode;;
  
  let name = "graphics mode"
  and plural_name = "graphics modes"
  ;;
  
  let default = default_graphics_mode;;
  
  let of_string = function
    | "no_graphics" | "ng" -> No_graphics
    | "minimal_graphics" | "mg" -> Minimal_graphics
    | "parameter_graphics" | "pg" -> Parameter_graphics
    | "compound_graphics" | "cg" -> Compound_graphics
    | "full_graphics" | "fg" -> Full_graphics
    | s ->
        o_failwith "graphics_mode_of_string"
          (Printf.sprintf "Unknown %s %S" name s)
  ;;
  
  let to_string = function
    | No_graphics -> "no_graphics"
    | Minimal_graphics -> "minimal_graphics"
    | Parameter_graphics -> "parameter_graphics"
    | Compound_graphics -> "compound_graphics"
    | Full_graphics -> "full_graphics"
  ;;
  
  let all =
    [ No_graphics;
      Minimal_graphics; Parameter_graphics;
      Compound_graphics ; Full_graphics; ]
  ;;
end
;;

module type GRAPHICAL_MODE =
    Ustring_command_option_spec.T with type t = graphics_mode
;;
module Graphics_mode : GRAPHICAL_MODE =
  Ustring_command_option.Make (Graphics_mode_structure)
;;
let get_graphics_mode = Graphics_mode.get
and graphics_mode_rco = Graphics_mode.rco
;;

type graphics_window_min_width = int
and graphics_window_min_height = int
and graphics_window_max_width = int
and graphics_window_max_height = int
;;

let graphics_window_min_width_option_name = "graphics window min width"
and graphics_window_min_height_option_name = "graphics window min height"
and graphics_window_max_width_option_name = "graphics window max width"
and graphics_window_max_height_option_name = "graphics window max height"
;;

let default_graphics_window_min_width = 800
and default_graphics_window_min_height = 600
and default_graphics_window_max_width = 1920
and default_graphics_window_max_height = 1080
;;

let { Uaref.get = get_graphics_window_min_width;
      set = set_graphics_window_min_width; } =
  Uaref.make_aref_single
    graphics_window_min_width_option_name
    default_graphics_window_min_width

and { Uaref.get = get_graphics_window_min_height;
      set = set_graphics_window_min_height; } =
  Uaref.make_aref_single
    graphics_window_min_height_option_name
    default_graphics_window_min_height

and { Uaref.get = get_graphics_window_max_width;
      set = set_graphics_window_max_width; } =
  Uaref.make_aref_single
    graphics_window_max_width_option_name
    default_graphics_window_max_width

and { Uaref.get = get_graphics_window_max_height;
      set = set_graphics_window_max_height; } =
  Uaref.make_aref_single
    graphics_window_max_height_option_name
    default_graphics_window_max_height
;;

let { long = graphics_window_min_width_lok;
      short = graphics_window_min_width_sok_st; } =
  make_option_keys
    graphics_window_min_width_option_name
    (Short (of_string "-g-min-w"))

and { long = graphics_window_min_height_lok;
      short = graphics_window_min_height_sok_st; } =
  make_option_keys
    graphics_window_min_height_option_name
    (Short (of_string "-g-min-h"))

and { long = graphics_window_max_width_lok;
      short = graphics_window_max_width_sok_st; } =
  make_option_keys
    graphics_window_max_width_option_name
    (Short (of_string "-g-max-w"))

and { long = graphics_window_max_height_lok;
      short = graphics_window_max_height_sok_st; } =
  make_option_keys
    graphics_window_max_height_option_name
    (Short (of_string "-g-max-h"))
;;

let graphics_window_min_width_rco =
  {
   Cmd_opt.key = graphics_window_min_width_lok;
   spec = Arg.Int set_graphics_window_min_width;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet graphics window minimum width to <nat>.\n\
        \tWidths are natural numbers.\n\
        \tDefault graphics window minimum width is %d."
       (get_graphics_window_min_width ());
  }

and graphics_window_min_height_rco =
  {
   Cmd_opt.key = graphics_window_min_height_lok;
   spec = Arg.Int set_graphics_window_min_height;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet graphics window minimum height to <nat>.\n\
        \tHeights are natural numbers.\n\
        \tDefault graphics window minimum height is %d."
       (get_graphics_window_min_height ());
  }

and graphics_window_max_width_rco =
  {
   Cmd_opt.key = graphics_window_max_width_lok;
   spec = Arg.Int set_graphics_window_max_width;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet graphics window maximum width to <nat>.\n\
        \tWidths are natural numbers.\n\
        \tDefault graphics window maximum width is %d."
       (get_graphics_window_max_width ());
  }

and graphics_window_max_height_rco =
  {
   Cmd_opt.key = graphics_window_max_height_lok;
   spec = Arg.Int set_graphics_window_max_height;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet graphics window maximum height to <nat>.\n\
        \tHeights are natural numbers.\n\
        \tDefault graphics window maximum height is %d."
       (get_graphics_window_max_height ());
  }
;;

type graphics_gradient_threshold = int;;

let graphics_gradient_threshold_option_name = "graphics gradient threshold"
and default_graphics_gradient_threshold = 10
;;

let { Uaref.get = get_graphics_gradient_threshold;
      set = set_graphics_gradient_threshold; } =
  Uaref.make_aref_single
    graphics_gradient_threshold_option_name
    default_graphics_gradient_threshold
;;

let { long = graphics_gradient_threshold_lok;
      short = graphics_gradient_threshold_sok_st; } =
  make_option_keys
    graphics_gradient_threshold_option_name
    (Short (of_string "-g-gt"))
;;

let graphics_gradient_threshold_rco =
  {
   Cmd_opt.key = graphics_gradient_threshold_lok;
   spec = Arg.Int set_graphics_gradient_threshold;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet graphics gradient threshold to <nat>.\n\
        \tThresholds are natural numbers.\n\
        \tDefault graphics gradient threshold is %d."
       (get_graphics_gradient_threshold ());
  }
;;

(* Initial and final images. *)

let init_initial_image_filename = "";;

let { Uaref.set = set_argument_initial_image_filename; _; } =
  Uaref.make_aref_unique
    "initial image file name" init_initial_image_filename
;;

type argument_initial_image = image
and argument_initial_image_width = width
and argument_initial_image_height = height
;;

let init_initial_image = Umatrix.of_array_array [||]
and init_initial_image_width = Nat._0_
and init_initial_image_height = Nat._0_
;;

let { Uaref.get = get_argument_initial_image;
      set = set_argument_initial_image; } =
  Uaref.make_aref_unique "initial image" init_initial_image
and { Uaref.get = get_argument_initial_image_width;
      set = set_argument_initial_image_width; } =
  Uaref.make_aref_unique "initial image width" init_initial_image_width
and { Uaref.get = get_argument_initial_image_height;
      set = set_argument_initial_image_height; } =
  Uaref.make_aref_unique "initial image height" init_initial_image_height
;;

let set_initial_image fname =
  set_argument_initial_image_filename fname;
  let pixmap = Ppm.load_pixmap fname in
  let idata = Umatrix.of_array_array pixmap in
  set_argument_initial_image idata;
  let (h, w) = Umatrix.size idata in
  if h = Nat._0_ || w = Nat._0_ then
    begin print_endline "empty image, nothing to do"; exit 0 end else
  begin set_argument_initial_image_width w;
  set_argument_initial_image_height h end;
;;

type final_image_filename = Ufilename.t;;

let final_image_filename_option_name = "final image filename"
and default_final_image_filename = "final_image.ppm"
;;

let { Uaref.get = get_final_image_filename;
      set = set_final_image_filename; } =
  Uaref.make_aref_single
    final_image_filename_option_name
    default_final_image_filename
;;

let { long = final_image_filename_lok;
      short = final_image_filename_sok_st; } =
  make_option_keys final_image_filename_option_name (Short (of_string "-o"))
;;

let final_image_filename_rco =
  {
   Cmd_opt.key = final_image_filename_lok;
   spec = Arg.String set_final_image_filename;
   doc =
     Printf.sprintf
       "<fname>\n\
        \tSet final image file name to <fname>.\n\
        \tDefault is %S."
       (get_final_image_filename ());
  }
;;

(* Report/debug flags and error messages. *)

let printing_facilities =
  Udebug.application_printing_facilities Ref_image_version.option_key_data
;;

(* Command-line options. *)

let phase_dimension = Cmd_opt.Phase "dimension";;

let add_options () =
  Cmd_opt.application_add_all_rco_with_sk
    [
     (codomain_implementation_sok_st, codomain_implementation_rco);
     (dimension_sok_st, dimension_rco);
     (dimension_management_sok_st, dimension_management_rco);
     (number_of_cuts_sok_st, number_of_cuts_rco);
     (clear_basic_cutting_families_sok_st, clear_basic_cutting_families_rco);
     (add_basic_cutting_family_sok_st, add_basic_cutting_family_rco);
     (cutting_strategy_sok_st, cutting_strategy_rco);
     (max_iteration_sok_st, max_iteration_rco);
     (min_best_scalar_indicator_sok_st, min_best_scalar_indicator_rco);
     (min_scalar_relative_error_sok_st, min_scalar_relative_error_rco);
     (out_formatter_sok_st, out_formatter_rco);
     (report_allocated_memory_sok_st, report_allocated_memory_rco);
     (incremental_mode_sok_st, incremental_mode_rco);
     (graphics_mode_sok_st, graphics_mode_rco);
     (graphics_window_min_width_sok_st, graphics_window_min_width_rco);
     (graphics_window_min_height_sok_st, graphics_window_min_height_rco);
     (graphics_window_max_width_sok_st, graphics_window_max_width_rco);
     (graphics_window_max_height_sok_st, graphics_window_max_height_rco);
     (graphics_gradient_threshold_sok_st, graphics_gradient_threshold_rco);
     (final_image_filename_sok_st, final_image_filename_rco);
    ];
  Cmd_opt.phase_add_all
    phase_dimension
    [ codomain_implementation_lok; dimension_lok; ];
;;
add_options ();;

Cmd_opt.set_anon_fun set_initial_image;;
Cmd_opt.set_usage_msg
  (Format.sprintf
     "@.%s -- Segmentation of images@.@.Options are:@."
     Sys.argv.(0))
;;
