(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uarray.mli,v 1.41 2017/10/30 18:38:49 fclement Exp $ *)

(** {3 More array operations}

 Length, index and offset arguments are naturals of type [Unatural.t].
 For sake of simplicity, natural constants and operations are represented
 here by their integer counterparts.
 Thus,  in the following, [0] should be read {!Unatural._0_}, and the infix
 operators [( + )], [( - )], [( * )], and [( / )] should be read respectively
 as the functions {!Unatural.add}, {!Unatural.sub}, {!Unatural.mul}, and
 {!Unatural.div}. *)

type 'a t = 'a array
(** Alias for type [array]. *)
;;

type length = Upervasives.length
and index = Upervasives.index
and offset = Upervasives.offset
(** Aliases for type {!Unatural.t}. *)
;;

val length : 'a t -> length
(** Return the length (number of elements) of the given array. *)
;;

val get : 'a t -> index -> 'a
(** [get a n] returns the element number [n] of array [a].
 The first element has number [0].
 The last element has number {!length} [a].

 Raise [Invalid_argument] if [n] is outside the range [0] to
 {!length} [a - 1]. *)
;;

val set : 'a t -> index -> 'a -> unit
(** [set a n x] modifies array [a] in place, replacing element number [n]
 with [x].

 Raise [Invalid_argument] if [n] is outside the range [0] to
 {!length} [a - 1]. *)
;;

val empty : 'a t
(** The empty array [\[||\]]. *)
;;

val make : length -> 'a -> 'a t
(** [make n x] returns a fresh array of length [n],
 initialized with [x].
 All the elements of this new array are initially
 physically equal to [x] (in the sense of the [==] predicate).
 Consequently, if [x] is mutable, it is shared among all elements
 of the array, and modifying [x] through one of the array entries
 will modify all other entries at the same time.

 Raise [Invalid_argument] if [n > Sys.max_array_length].
 If the value of [x] is a floating-point number, then the maximum
 size is only [Sys.max_array_length / 2].*)
;;

val init : length -> (index -> 'a) -> 'a t
(** [init n f] returns a fresh array of length [n],
 with element number [i] initialized to the result of [f i].
 In other terms, [init n f] tabulates the results of [f]
 applied to the integers [0] to [n-1].

 Raise [Invalid_argument] if [n > Sys.max_array_length].
 If the return type of [f] is [float], then the maximum
 size is only [Sys.max_array_length / 2].*)
;;

val init_rec : length -> ('a t -> index -> 'a) -> 'a t;;

type 'a init_rec1 = 'a
and 'a prev_rec1 = 'a
;;

val init_rec1 :
  length -> 'a init_rec1 -> ('a prev_rec1 -> index -> 'a) -> 'a t
;;

type 'a scheme_rec1_lin = 'a;;
type 'a prod_scal1 = 'a scheme_rec1_lin -> 'a prev_rec1 -> 'a;;

val init_rec1_lin :
  length -> 'a init_rec1 -> 'a scheme_rec1_lin -> 'a prod_scal1 -> 'a t
;;

type 'a init_rec2 = 'a * 'a
and 'a prev_rec2 = 'a * 'a
;;

val init_rec2 :
  length -> 'a init_rec2 -> ('a prev_rec2 -> index -> 'a) -> 'a t
;;

type 'a scheme_rec2_lin = 'a * 'a;;
type 'a prod_scal2 = 'a scheme_rec2_lin -> 'a prev_rec2 -> 'a;;

val init_rec2_lin :
  length -> 'a init_rec2 -> 'a scheme_rec2_lin -> 'a prod_scal2 -> 'a t
;;

val append : 'a t -> 'a t -> 'a t
(** [append a1 a2] returns a fresh array containing the concatenation of the
 arrays [a1] and [a2]. *)
;;

val concat : 'a t list -> 'a t
(** Same as {!append}, but concatenates a list of arrays. *)
;;

val sub : 'a t -> offset -> length -> 'a t
(** [sub a ofs len] returns a fresh array of length [len], containing the
 elements number [ofs] to [ofs + len - 1] of array [a].

 Raise [Invalid_argument] if the subarray specification is wrong; that is,
 if [ofs + len > length a]. *)
;;

val copy : 'a t -> 'a t
(** [copy a] returns a copy of [a], that is, a fresh array containing the
 same elements as [a]. *)
;;

val is_empty : 'a t -> bool
(** [is_empty a] is true if and only if the array [a] is empty. *)
;;

val check_nonempty : Upervasives.function_name -> 'a t -> unit
(** Check if the given array is nonempty.

 Raise [Invalid_argument] when the array is empty. *)
;;

val get_same :
  Upervasives.qualify -> Upervasives.function_name ->
  (index -> 'b -> 'b -> Upervasives.message) ->
  ('a -> 'b) -> 'a t -> 'b
(** [get_same qual fun_name gen_msg f a] returns the value of [f] applied to
 elements of array [a] when it is constant; otherwise raises
 [Failure] when an element produces a different result.
 Let [i] be the index of this element ([i > 0]).
 Then, the string associated with the exception is computed as
 [gen_msg i (f a.(i)) (f a.(0))], and it is prefixed using [qual] and
 [fun_name] (see {!Upervasives.invalid_arg_qual} for their use).

 Raise [Invalid_argument] when the array [a] is empty. *)
;;

(** {6 Filling arrays} *)

val sub_fill : ('a -> 'a) -> 'a t -> offset -> length -> unit
(** [sub_fill f a ofs len] applies function [f] in turn to the
 elements number [ofs] to [ofs + len - 1] of array [a] and stores the result
 to the corresponding location. It is equivalent to
 [a.(ofs) <- f a.(ofs);
 a.(ofs + 1) <- f a.(ofs + 1);
 ...;
 a.(ofs + len - 1) <- f a.(ofs + len - 1)].

 Raise [Invalid_argument] if the subarray specification is wrong
 (see {!sub}). *)
;;

val fill : ('a -> 'a) -> 'a t -> unit
(** [fill f a] is equivalent to {!sub_fill} [f a 0 (]{!length} [a)]. *)
;;

val sub_fill_i :
 (index -> 'a -> 'a) -> 'a t -> offset -> length -> unit
(** Same as {!sub_fill}, but the function is applied to the index
 of the element as first argument, and the element itself as second argument.
 *)
;;

val fill_i : (index -> 'a -> 'a) -> 'a t -> unit
(** [fill_i f a] is equivalent to {!sub_fill_i} [f a 0 (]{!length} [a)]. *)
;;

val blit : 'a t -> offset -> 'a t -> offset -> length -> unit
(** [blit a1 ofs1 a2 ofs2 len] copies [len] elements from array [a1], starting
 at element number [ofs1], to array [a2], starting at element number [ofs2].
 It works correctly even if [a1] and [a2] are the same array, and the source
 and destination chunks overlap.

 Raise [Invalid_argument] if the subarray specification is wrong
 (see {!sub}). *)
;;

(** {6 Converters} *)

val to_list : 'a t -> 'a list
(** [to_list a] returns the list of all the elements of [a]. *)
;;

val of_list : 'a list -> 'a t
(** [of_list l] returns a fresh array containing the elements of [l]. *)
;;

val list_of_list_array : 'a list t -> 'a t list;;

val list_to_list_array : 'a t list -> 'a list t;;

val of_list1 : 'a Ulist1.t -> 'a t;;

(** {6 Iterators} *)

type index_min = index
and index_max = index
;;

val nat_for_loop : index_min -> index_max -> (index -> unit) -> unit
(** Alias for {!Unatural.for_loop}. *)
;;

val nat_rev_for_loop : index_max -> index_min -> (index -> unit) -> unit
(** Alias for {!Unatural.for_loop}. *)
;;

val for_loop : offset -> length -> (index -> unit) -> unit
(** [for_loop ofs len f] is equivalent to
 {!nat_for_loop} [ofs (ofs + len - 1) f]. *)
;;

val rev_for_loop : offset -> length -> (index -> unit) -> unit
(** [rev_for_loop ofs len f] is equivalent to
 {!nat_rev_for_loop} [(ofs + len - 1) ofs f]. *)
;;

val sub_iter : ('a -> unit) -> 'a t -> offset -> length -> unit
(** [sub_iter] is equivalent to [iter] for the specified subarray.

 Raise [Invalid_argument] if the subarray specification is wrong
 (see {!sub}). *)
;;

val sub_iter_i : (index -> 'a -> unit) -> 'a t -> offset -> length -> unit
(** Same as {!sub_iter}, but the function is applied to the index of
 the element as first argument, and the element itself as second
 argument. *)
;;

val iter : ('a -> unit) -> 'a t -> unit
(** [iter f a] is equivalent to [sub_iter f a 0 (]{!length} [a)]. *)
;;

val iter_i : (index -> 'a -> unit) -> 'a t -> unit
(** [iter_i f a] is equivalent to [sub_iter_i f a 0 (]{!length} [a)]. *)
;;

val iter_rails : ('a -> unit) t -> 'a t -> unit
(** [iter_rails fa a] applies components of function array [fa] in turn to
 all the elements of [a].
 It is equivalent to
 [fa.(0) a.(0); fa.(1) a.(1); ...; fa.(n - 1) a.(n - 1); ()]
 where [n] is the common length of arrays [fa] and [a].

 Raise [Invalid_argument] if the two arrays have different lengths. *)
;;

val map : ('a -> 'b) -> 'a t -> 'b t
(** [map f a] applies function [f] to all the elements of [a], and builds an
 array with the results returned by [f]:
 [\[| f a.(0); f a.(1); ...; f a.(]{!length} [a - 1) |\]]. *)
;;

val map_i : (index -> 'a -> 'b) -> 'a t -> 'b t
(** Same as {!map}, but the function is applied to the index of the element
 as first argument, and the element itself as second argument. *)
;;

val rails : ('a -> 'b) t -> 'a t -> 'b t
(** [rails fa a] applies components of function array [fa] in turn to all the
 elements of [a], and builds an array with the results returned by [f].
 It is equivalent to
 [\[| fa.(0) a.(0); fa.(1) a.(1); ...; f.(n - 1) a.(n - 1); |\]]
 where [n] is the common length of arrays [fa] and [a].

 Raise [Invalid_argument] if the two arrays have different lengths. *)
;;

val fold_left : ('accu -> 'a -> 'accu) -> 'accu -> 'a t -> 'accu
(** [fold_left f x a] computes
 [f (... (f (f x a.(0)) a.(1)) ...) a.(n - 1)]
 where [n] is the length of the array [a]. *)
;;

val fold_right : ('a -> 'accu -> 'accu) -> 'a t -> 'accu -> 'accu
(** [fold_right f a x] computes
 [f a.(0) (f a.(1) (... (f a.(n - 1) x) ...))]
 where [n] is the length of the array [a]. *)
;;

val fold_left_i :
 ('accu -> index -> 'a -> 'accu) -> 'accu -> 'a t -> 'accu
(** Same as {!fold_left}, but the function is applied to the index of the
 element as second argument. *)
;;

val fold_right_i :
 (index -> 'b -> 'accu -> 'accu) -> 'b t -> 'accu -> 'accu
(** Same as {!fold_right}, but the function is applied to the index of the
 element as first argument. *)
;;

val sub_fold_left :
 ('accu -> 'a -> 'accu) -> 'accu -> 'a t -> offset -> length -> 'accu
(** [sub_fold_left] is equivalent to {!fold_left} for the specified subarray.

 Raise [Invalid_argument] if the subarray specification is wrong
 (see {!sub}). *)
;;

val sub_fold_right :
 ('a -> 'accu -> 'accu) -> 'a t -> offset -> length -> 'accu -> 'accu
(** [sub_fold_right] is equivalent to {!fold_right} for the specified
 subarray.

 Raise [Invalid_argument] if the subarray specification is wrong
 (see {!sub}). *)
;;

val sub_fold_left_i :
 ('accu -> index -> 'a -> 'accu) ->
 'accu -> 'a t -> offset -> length -> 'accu
(** Same as {!sub_fold_left}, but the function is applied to the index of the
 element as second additional argument. *)
;;

val sub_fold_right_i :
 (index -> 'a -> 'accu -> 'accu) ->
 'a t -> offset -> length -> 'accu -> 'accu
(** Same as {!sub_fold_right}, but the function is applied to the index of the
 element as first additional argument. *)
;;

(** {6 Iterators on two arrays} *)

val iter2 : ('a -> 'b -> unit) -> 'a t -> 'b t -> unit
(** [iter2 f a b] applies function [f] in turn to all the elements
 of [a] and [b]. It is equivalent to
 [f a.(0) b.(0); f a.(1) b.(1); ...; f a.(n - 1) b.(n - 1); ()],
 where [n] is the length of arrays [a] and [b].

 Raise [Invalid_argument] if the two arrays have different lengths. *)
;;

val map2 : ('a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t
(** [map2 f a b] applies function [f] to all the elements of [a]
 and [b], and builds an array with the results returned by [f]:
 [[| f a.(0) b.(0); f a.(1) b.(1); ...; f a.(n - 1) b.(n - 1) |]],
 where [n] is the length of arrays [a] and [b].

 Raise [Invalid_argument] if the two arrays have different lengths. *)
;;

val iter2_i : (index -> 'a -> 'b -> unit) -> 'a t -> 'b t -> unit
(** Same as {!iter2}, but the function is applied to the index of the
 elements as first argument, and the elements themselves as second and third
 arguments. *)
;;

val map2_i : (index -> 'a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t
(** Same as {!map2}, but the function is applied to the index of the
 elements as first argument, and the elements themselves as second and third
 arguments. *)
;;

val fold_left2 :
  ('accu -> 'a -> 'b -> 'accu) -> 'accu -> 'a t -> 'b t -> 'accu
(** [fold_left2 f x a b] computes
 [f (... (f (f x a.(0) b.(0)) a.(1) b.(1)) ...) a.(n - 1) b.(n - 1)],
 where [n] is the length of the arrays [a] and [b].

 Raise [Invalid_argument] if the two arrays have different lengths. *)
;;

val fold_right2 :
  ('a -> 'b -> 'accu -> 'accu) -> 'a t -> 'b t -> 'accu -> 'accu
(** [fold_right2 f a b x] computes
 [f a.(0) b.(0) (f a.(1) b.(1) ( ... (f a.(n - 1) b.(n - 1) x) ...))],
 where [n] is the length of the arrays [a] and [b].

 Raise [Invalid_argument] if the two arrays have different lengths. *)
;;

val sub_iter2 :
  ('a -> 'b -> unit) -> 'a t -> offset -> 'b t -> offset -> length -> unit
(** [sub_iter2] is equivalent to {!iter2} for the specified subarray.

 Raise [Invalid_argument] if the subarray specification is wrong. *)
;;

val sub_iter2_i :
  (index -> 'a -> index -> 'b -> unit) ->
  'a t -> offset -> 'b t -> offset -> length -> unit
(** Same as {!sub_iter2}, but the function is applied to the index of
 the elements as first and third arguments. *)
;;

(** {6 Iterators on more arrays} *)

val iter3 : ('a -> 'b -> 'c -> unit) -> 'a t -> 'b t -> 'c t -> unit;;
val iter4 :
  ('a -> 'b -> 'c -> 'd -> unit) -> 'a t -> 'b t -> 'c t -> 'd t -> unit
;;
val map3 : ('a -> 'b -> 'c -> 'd) -> 'a t -> 'b t -> 'c t -> 'd t;;
val map4 :
  ('a -> 'b -> 'c -> 'd -> 'e) -> 'a t -> 'b t -> 'c t -> 'd t -> 'e t
(** See {!iter2} and {!map2}.

 Raise [Invalid_argument] if argument arrays have different lengths. *)
;;

(** {6 Iterators on list of arrays} *)

val map_list : ('a list -> 'a) -> 'a t list -> 'a t
(** [map_list f al] applies function [f] to all the elements of the
 items of the array list [al], and builds an array with the results returned
 by [f]:
 [[| f la.(0); f la.(1); ...; f la.(n - 1) |]],
 where [la] is the array of lists [[| [a1.(0); a2.(0); ...; ap.(0)];
 [a1.(1); ...; ap.(1)]; ...;[a1.(n - 1); ...; ap.(n - 1)] |]]
 and [n] is the length of arrays of the list [al].

 Raise [Invalid_argument] if the arrays of the list have different
 lengths. *)
;;

(** {6 Array scanning} *)

val for_all : ('a -> bool) -> 'a t -> bool
(** [for_all p [| a1; ...; an; |]] checks if all elements of the array
 satisfy the predicate [p]. That is, it returns
 [p a1 && p a2 && ... && p an]. *)
;;

val exists : ('a -> bool) -> 'a t -> bool
(** [exists p [| a1; ...; an; |]] checks if at least one element of
 the array satisfies the predicate [p]. That is, it returns
 [p a1 || p a2 || ... || p an]. *)
;;

val mem : 'a -> 'a t -> bool
(** [mem x a] is true if and only if [x] is equal to an element of [a]. *)
;;

val memq : 'a -> 'a t -> bool
(** Same as {!mem}, but uses physical equality instead of structural equality
 to compare array elements. *)
;;

val find : ('a -> bool) -> 'a t -> 'a
(** [find p [|a1; ...; an|]] returns the first element of the array
 that satisfies the predicat [p] if any, otherwise raises [Not_found]. *)
;;

val rev_find_all : ('a -> bool) -> 'a t -> 'a list
(** [rev_find_all p [|a1; ...; an|]] returns all the elements of the array
 that satisfies the predicat [p] if any (in reverse order), otherwise
 returns [[]]. *)
;;

val rev_find_all_i : (index -> 'a -> bool) -> 'a t -> 'a list
(** Same as [rev_find_all] but the predicate argument takes the index of the
 element as first argument. *)
;;

val for_all2 : ('a -> 'b -> bool) -> 'a t -> 'b t -> bool;;

val sub_find : ('a -> bool) -> 'a t -> offset -> length -> 'a
(** [sub_find] is equivalent to {!find} for the specified subarray.

 Raise [Invalid_argument] if the subarray specification is wrong. *)
;;

val rev_sub_find_all : ('a -> bool) -> 'a t -> offset -> length -> 'a list
(** [rev_sub_find_all] is equivalent to {!rev_find_all} for the specified
 subarray.

 Raise [Invalid_argument] if the subarray specification is wrong. *)
;;

val rev_sub_find_all_i :
  (index -> 'a -> bool) -> 'a t -> offset -> length -> 'a list
(** Same as [rev_sub_find_all] but the predicate argument takes the index of
 the element as first argument. *)
;;

(** {6 Selecting elements in arrays} *)

val select : Upervasives.select t -> 'a t -> 'a t
(** [select chi a] returns a fresh array containing elements [a.(i)]
 such that [chi.(i)] is {!Upervasives}.[Keep]. The order is preserved.

 Raise [Invalid_argument] if the two arrays have different lengths. *)
;;

val map_select : ('a -> 'b) -> Upervasives.select t -> 'a t -> 'b t
(** [map_select f chi a] returns a fresh array containing elements [f a.(i)]
 such that [chi.(i)] is {!Upervasives}.[Keep]. The order is preserved.

 Raise [Invalid_argument] if the two arrays have different lengths. *)
;;

val unselect : 'a -> Upervasives.select t -> 'a t -> 'a t
(** [unselect x chi a] returns a fresh array of the same length as [chi]
 such that its [i]-th element is [x] when [chi.(i) =] {!Upervasives}[.Forget]
 and [a.(j)] when [chi.(i)] is the [j]-th {!Upervasives}[.Keep] value.

 For all [x], [select chi (unselect x chi a)] gives back an array equal to
 [a].

 Raise [Failure] if the number of {!Upervasives}[.Keep] values in [chi] is
 different from the length of [a]. *)
;;

(** {6 Sorting arrays} *)

val sort : ('a -> 'a -> int) -> 'a t -> unit
(** See [Array.sort]. *)
;;

val stable_sort : ('a -> 'a -> int) -> 'a t -> unit
(** See {!sort}. *)
;;

val fast_sort : ('a -> 'a -> int) -> 'a t -> unit
(** See {!sort}. *)
;;

val is_sorted : ('a -> 'a -> bool) -> 'a t -> bool;;

val is_strictly_increasing : 'a t -> bool;;
val check_strictly_increasing :
 Upervasives.function_name -> 'a t -> unit
;;

val sorted_array_find : 'a t -> 'a -> index;;
val sorted_array_findq : 'a t -> 'a -> index;;

(**/**)
(** {6 Undocumented features} *)

val unsafe_get : 'a t -> index -> 'a;;
val unsafe_set : 'a t -> index -> 'a -> unit;;
