(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: udebug.ml,v 1.55 2017/05/31 15:24:06 fclement Exp $ *)

(* Debug/report/error message facilities. *)

module Ar = Uaref;;
module K = Uoption_key;;
module Co = Ucommand_option;;

type tag = string
and 'a make_tag_opt = 'a -> tag option
;;

type 'a print = 'a -> unit
and 'a gen_print = Upervasives.function_name option -> 'a print
;;

type message_print = Upervasives.message print
and gen_message_print = Upervasives.gen_message print
and message_gen_print = Upervasives.message gen_print
and gen_message_gen_print = Upervasives.gen_message gen_print
;;

type printing_facilities =
  {
   get_warn_error : unit -> bool;
   report : message_print;
   debug : message_gen_print;
   warning : message_gen_print;
   error : message_gen_print;
   report_action : gen_message_print;
   debug_action : gen_message_gen_print;
   warning_action : gen_message_gen_print;
   error_action : gen_message_gen_print;
  }

and 'a dynamic_printing_facilities =
  {
   dyn_get_warn_error : unit -> bool;
   dyn_report : 'a -> message_print;
   dyn_debug : 'a -> message_gen_print;
   dyn_warning : 'a -> message_gen_print;
   dyn_error : 'a -> message_gen_print;
   dyn_report_action : 'a -> gen_message_print;
   dyn_debug_action : 'a -> gen_message_gen_print;
   dyn_warning_action : 'a -> gen_message_gen_print;
   dyn_error_action : 'a -> gen_message_gen_print;
  }
;;

(* Printing. *)

let display_message_on_stderr ob msg =
  Printf.bprintf ob "%s\n" msg;
  Ubuffer.flush_buffer_on_stderr ob
;;

(* Messages with dynamic tags. *)

type option_kind =
  | Warn_error
  | Report
  | Debug
;;

let option_class_of_option_king = function
  | Warn_error -> Co.Warn_error ()
  | Report -> Co.Report ()
  | Debug -> Co.Debug ()
;;

let string_of_option_kind ok =
  Co.string_of_option_kind (option_class_of_option_king ok)
and keys_of_option_kind ok =
  Co.option_keys_of_option_kind (option_class_of_option_king ok)
;;

let print_sname () { Co.name = sname; _; } = Printf.sprintf " %s" sname;;

let printing_facilities_with_dynamic_tags purpose sk_data make_tag_opt =

  (*  Warn error, report, and debug flags. *)

  let warn_error_flag = Ar.make_flag ()
  and (report_flag, debug_flag) = Ar.make_embedded_flags () in

  let add_flag ok comment set_flag add =
    let kind = string_of_option_kind ok in
    let keys = keys_of_option_kind ok in
    let { K.long = lk; short = sk_st; } =
      match sk_data.Co.option_key_prefixes_opt with
      | None -> keys
      | Some kps -> K.concat_option_keys kps keys in
    let short_option_key =
      match sk_st with
      | K.Automatic -> Co.with_auto_short_option_key
      | K.Short sk -> Co.with_short_option_key sk in
    add
      (short_option_key
         (Co.with_auto_gnu_option_key
            (Co.of_regular
               { Co.key = lk;
                 spec = Arg.Unit set_flag;
                 doc =
                   Printf.sprintf
                     "\n\tTurn%a %s mode on. Default is off.%s"
                     print_sname sk_data kind comment; }))) in

  let add_warn_error_flag () =
    let add =
      match purpose with
      | Co.Application -> Co.warn_error_app_add
      | Co.Library -> Co.warn_error_lib_add in
    add_flag Warn_error "" warn_error_flag.Ar.set_flag add in
  add_warn_error_flag ();

  let add_report_flag () =
    let add =
      match purpose with
      | Co.Application -> Co.report_app_add
      | Co.Library -> Co.report_lib_add in
    add_flag Report "" report_flag.Ar.set_flag add in
  add_report_flag ();

  let add_debug_flag () =
    let add =
      match purpose with
      | Co.Application -> Co.debug_app_add
      | Co.Library -> Co.debug_lib_add in
    let cmt =
      Printf.sprintf "\n\tImplies%a report mode." print_sname sk_data in
    add_flag Debug cmt debug_flag.Ar.set_flag add in
  add_debug_flag ();

  (* Printing functions. *)

  let ob = Buffer.create 1024 in

  let report_action x action =
    if report_flag.Ar.get_flag () then
      let tag_opt = make_tag_opt x in
      let msg = Upervasives.report_message tag_opt (action ()) in
      display_message_on_stderr ob msg in

  let report x msg = report_action x (fun () -> msg) in

  let gen_action get_flag gen_message_opt fun_name_opt x action =
    if get_flag () then
      let tag_opt = make_tag_opt x in
      let msg = gen_message_opt tag_opt fun_name_opt (action ()) in
      display_message_on_stderr ob msg in

  let debug_action x fun_name_opt action =
    let get_flag = debug_flag.Ar.get_flag in
    gen_action get_flag Upervasives.debug_message fun_name_opt x action in

  let debug x fun_name_opt msg =
    debug_action x fun_name_opt (fun () -> msg) in

  let warning_action x fun_name_opt action =
    let get_flag () = true in
    gen_action get_flag Upervasives.warning_message fun_name_opt x action in

  let warning x fun_name_opt msg =
    warning_action x fun_name_opt (fun () -> msg) in

  let error_action x fun_name_opt action =
    let get_flag () = true in
    gen_action get_flag Upervasives.error_message fun_name_opt x action in

  let error x fun_name_opt msg =
    error_action x fun_name_opt (fun () -> msg) in

  {
   dyn_get_warn_error = warn_error_flag.Ar.get_flag;
   dyn_report = report;
   dyn_debug = debug;
   dyn_warning = warning;
   dyn_error = error;
   dyn_report_action = report_action;
   dyn_debug_action = debug_action;
   dyn_warning_action = warning_action;
   dyn_error_action = error_action;
  }
;;

let application_printing_facilities_with_dynamic_tags sname_opt =
  printing_facilities_with_dynamic_tags Co.Application sname_opt
and library_printing_facilities_with_dynamic_tags sname_opt =
  printing_facilities_with_dynamic_tags Co.Library sname_opt
;;

let gen_printing_facilities purpose sk_data make_tag_opt =
  let
      {
       dyn_get_warn_error = get_warn_error;
       dyn_report = report;
       dyn_debug = debug;
       dyn_warning = warning;
       dyn_error = error;
       dyn_report_action = report_action;
       dyn_debug_action = debug_action;
       dyn_warning_action = warning_action;
       dyn_error_action = error_action;
      } =
    printing_facilities_with_dynamic_tags purpose sk_data make_tag_opt in
  {
   get_warn_error = get_warn_error;
   report = report ();
   debug = debug ();
   warning = warning ();
   error = error ();
   report_action = report_action ();
   debug_action = debug_action ();
   warning_action = warning_action ();
   error_action = error_action ();
  }
;;

let printing_facilities purpose sk_data =
  let make_tag_opt () = None in
  gen_printing_facilities purpose sk_data make_tag_opt
;;

let application_printing_facilities = printing_facilities Co.Application
and library_printing_facilities = printing_facilities Co.Library
;;

let printing_facilities_with_static_tag purpose sk_data =
    let tag = Printf.sprintf "[%s]" sk_data.Co.name in
    let make_tag_opt () = Some tag in
    gen_printing_facilities purpose sk_data make_tag_opt
;;

let application_printing_facilities_with_static_tag =
  printing_facilities_with_static_tag Co.Application
and library_printing_facilities_with_static_tag =
  printing_facilities_with_static_tag Co.Library
;;

(* Specializing printed messages. *)

type specialized_printing_facilities =
  {
   spf_get_warn_error : unit -> bool;
   spf_report : message_print;
   spf_debug : message_print;
   spf_warning : message_print;
   spf_error : message_print;
   spf_report_action : gen_message_print;
   spf_debug_action : gen_message_print;
   spf_warning_action : gen_message_print;
   spf_error_action : gen_message_print;
  }
;;

let set_function_name fun_name =
  let set_fun_name print = print (Some fun_name) in
  fun pf ->
    {
     spf_get_warn_error = pf.get_warn_error;
     spf_report = pf.report;
     spf_debug = set_fun_name pf.debug;
     spf_warning = set_fun_name pf.warning;
     spf_error = set_fun_name pf.error;
     spf_report_action = pf.report_action;
     spf_debug_action = set_fun_name pf.debug_action;
     spf_warning_action = set_fun_name pf.warning_action;
     spf_error_action = set_fun_name pf.error_action;
    }
;;

let prefix_messages pref =
  let lift print msg = print (Upervasives.prefix pref msg) in
  let lift_action print_action action =
    print_action (fun () -> Upervasives.prefix pref (action ())) in
  fun spf ->
    {
     spf with
     spf_report = lift spf.spf_report;
     spf_debug = lift spf.spf_debug;
     spf_warning = lift spf.spf_warning;
     spf_error = lift spf.spf_error;
     spf_report_action = lift_action spf.spf_report_action;
     spf_debug_action = lift_action spf.spf_debug_action;
     spf_warning_action = lift_action spf.spf_warning_action;
     spf_error_action = lift_action spf.spf_error_action;
    }

type 'a specialized_dynamic_printing_facilities =
  {
   sdpf_get_warn_error : unit -> bool;
   sdpf_report : 'a -> message_print;
   sdpf_debug : 'a -> message_print;
   sdpf_warning : 'a -> message_print;
   sdpf_error : 'a -> message_print;
   sdpf_report_action : 'a -> gen_message_print;
   sdpf_debug_action : 'a -> gen_message_print;
   sdpf_warning_action : 'a -> gen_message_print;
   sdpf_error_action : 'a -> gen_message_print;
  }
;;

let set_function_name_dyn fun_name =
  let set_fun_name dyn_print x = dyn_print x (Some fun_name) in
  fun dpf ->
    {
     sdpf_get_warn_error = dpf.dyn_get_warn_error;
     sdpf_report = dpf.dyn_report;
     sdpf_debug = set_fun_name dpf.dyn_debug;
     sdpf_warning = set_fun_name dpf.dyn_warning;
     sdpf_error = set_fun_name dpf.dyn_error;
     sdpf_report_action = dpf.dyn_report_action;
     sdpf_debug_action = set_fun_name dpf.dyn_debug_action;
     sdpf_warning_action = set_fun_name dpf.dyn_warning_action;
     sdpf_error_action = set_fun_name dpf.dyn_error_action;
    }
;;

let prefix_messages_dyn pref =
  let lift dyn_print x msg = dyn_print x (Upervasives.prefix pref msg) in
  let lift_action dyn_print_action x action =
    dyn_print_action x (fun () -> Upervasives.prefix pref (action ())) in
  fun sdpf ->
    {
     sdpf with
     sdpf_report = lift sdpf.sdpf_report;
     sdpf_debug = lift sdpf.sdpf_debug;
     sdpf_warning = lift sdpf.sdpf_warning;
     sdpf_error = lift sdpf.sdpf_error;
     sdpf_report_action = lift_action sdpf.sdpf_report_action;
     sdpf_debug_action = lift_action sdpf.sdpf_debug_action;
     sdpf_warning_action = lift_action sdpf.sdpf_warning_action;
     sdpf_error_action = lift_action sdpf.sdpf_error_action;
    }
;;

(* Managing automatic service options. *)

let warn_error_app_clear = Co.warn_error_app_clear
and warn_error_lib_clear = Co.warn_error_lib_clear
and report_app_clear = Co.report_app_clear
and report_lib_clear = Co.report_lib_clear
and debug_app_clear = Co.debug_app_clear
and debug_lib_clear = Co.debug_lib_clear
;;

let warn_error_app_clear_option_key = Co.warn_error_app_clear_option_key
and warn_error_lib_clear_option_key = Co.warn_error_lib_clear_option_key
and report_app_clear_option_key = Co.report_app_clear_option_key
and report_lib_clear_option_key = Co.report_lib_clear_option_key
and debug_app_clear_option_key = Co.debug_app_clear_option_key
and debug_lib_clear_option_key = Co.debug_lib_clear_option_key
;;

let phase_ulib_debugging = Co.phase_ulib_debugging
and phase_warn_error = Co.phase_warn_error
and phase_report = Co.phase_report
and phase_debug = Co.phase_debug
and phase_all_debugging = Co.phase_all_debugging
;;
