(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: udynamic_array.ml,v 1.14 2017/05/31 15:24:06 fclement Exp $ *)

(* Dynamic arrays. *)

module Nat = Unatural;;
module A = Uarray;;

let _0_ = Nat._0_
and _2_ = Nat._2_
;;

let qual = Upervasives.qualify "Udynamic_array";;
let failwith = Upervasives.failwith_qual qual;;

type index = Nat.t;;
type length = Nat.t;;

type 'a t =
  {
   mutable array : 'a array;
   default : 'a;
  }
and 'a dynamic_array = 'a t
;;

let same da1 da2 = da1 == da2;;

let equal_items equal_item da1 da2 =
  A.for_all2 equal_item da1.array da2.array
;;

let max_length = Nat.of_int (Sys.max_array_length / 2);;

let length da = A.length da.array;;

let make n x = { array = A.make n x; default = x; };;

let get da i = A.get da.array i;;

let grow fun_name da i =
  let len = length da in
  if i >= len then
    let new_len = max (Nat.succ i) (Nat.mul _2_ len) in
    let new_len = if new_len > max_length then max_length else new_len in
    if new_len < i then
      failwith fun_name "cannot create such a large dynamic array" else
    let new_a = A.make new_len da.default in
    A.blit da.array _0_ new_a _0_ len;
    da.array <- new_a
;;

let set da i x =
  grow "set" da i;
  A.set da.array i x
;;

(* Converters. *)

let to_array da = da.array;;
let of_array x a = { array = a; default = x; };;

(* Iterators. *)

let iter f da = A.iter_i f da.array;;

let map f da =
  {
   array = A.map_i f da.array;
   default = f _0_ da.default;
  }
;;

let fold_left f accu da =
  let ff accu i = f accu i in
  A.fold_left_i ff accu da.array
;;
