(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ulist.ml,v 1.40 2017/05/31 15:24:06 fclement Exp $ *)

(* List operations. *)

module L = List;;

module Nat = Unatural;;

let _0_ = Nat._0_;;

let qual = Upervasives.qualify "Ulist";;
let invalid_arg = Upervasives.invalid_arg_qual qual
let failwith = Upervasives.failwith_qual qual
;;

type length = Upervasives.length
and index = Upervasives.index
;;

let min_length = _0_;;

type 'a listpn = 'a list
and 'a t = 'a list
and 'a listn = 'a t
;;

let to_listpn = Upervasives.id
and of_listpn = Upervasives.id
and to_list = Upervasives.id
and of_list = Upervasives.id
and lift = Upervasives.id
and lift2 = Upervasives.id
;;

let cons x l = x :: l
and hd = L.hd
and tl = L.tl
and rev = L.rev
and append = L.append
and rev_append = L.rev_append
and iter = L.iter
and map = L.map
and rev_map = L.rev_map
and fold_left = L.fold_left
and fold_right = L.fold_right
and iter2 = L.iter2
and map2 = L.map2
and rev_map2 = L.rev_map2
and fold_left2 = L.fold_left2
and fold_right2 = L.fold_right2
and for_all = L.for_all
and exists = L.exists
and for_all2 = L.for_all2
and exists2 = L.exists2
and mem = L.mem
and memq = L.memq
and find = L.find
and filter = L.filter
and find_all = L.find_all
and partition = L.partition
and assoc = L.assoc
and assq = L.assq
and mem_assoc = L.mem_assoc
and mem_assq = L.mem_assq
and remove_assoc = L.remove_assoc
and remove_assq = L.remove_assq
and split = L.split
and combine = L.combine
and sort = L.sort
and stable_sort = L.stable_sort
and fast_sort = L.fast_sort
and merge = L.merge
;;

let length l = Nat.unsafe_of_int (L.length l);;

let nth l n = L.nth l (Nat.to_int n);;

let count n l =
  let rec loop i accu l =
    (* i is decreasing from n. *)
    if i = _0_ then accu else
    (* Now, we have i > 0.
       Therefore, the following Nat.pred will not fail. *)
    match l with
    | [] -> accu
    | x :: rest -> loop (Nat.pred i) (x :: accu) rest in
  loop n [] l
;;

let first_p n l = rev (count n l);;

let last_p n l = count n (rev l);;

let rec last = function
  | [] -> failwith "last" "empty list"
  | x :: [] -> x
  | _ :: (_ :: _ as l) -> last l
;;

let rec rank e = function
  | [] -> raise Not_found
  | x :: l -> if x = e then _0_ else Nat.succ (rank e l)
;;

let gen_init start next last f =
  let rec loop accu i =
    (* i goes from start to last. *)
    if i = last then f i :: accu else
    (* Now, we have i <> last.
       Therefore, if next = Nat.pred and last = 0, we have i > 0,
       and the following Nat.pred will not fail. *)
    loop (f i :: accu) (next i) in
  loop [] start
;;

let init n f =
  if n = _0_ then [] else
  (* Now, we have n > 0.
     Therefore, the following Nat.pred will not fail.
     See gen_init for a proof of safety for the second Nat.pred. *)
  gen_init (Nat.pred n) Nat.pred _0_ f
;;

let rev_init n f =
  if n = _0_ then [] else
  (* Now, we have n > 0.
     Therefore, the following Nat.pred will not fail. *)
  gen_init _0_ Nat.succ (Nat.pred n) f
;;

let replace_by_list i l1 l2 =
  let rec loop i2 = function
    (* i2 is increasing from 0. *)
    | [] -> []
    | x :: l ->
      if i2 <> i then x :: loop (Nat.succ i2) l else
      l1 @ loop (Nat.succ i2) l in
  loop _0_ l2
;;

let rev_replace_by_list i l1 l2 =
  let rec loop accu i2 = function
    (* i2 is increasing from 0. *)
    | [] -> accu
    | x :: l ->
      let accu = if i2 <> i then x :: accu else rev_append l1 accu in
      loop accu (Nat.succ i2) l in
  loop [] _0_ l2
;;

let insert i l1 l2 =
  let rec loop i2 = function
    (* i2 is increasing from 0. *)
    | [] -> []
    | x :: l ->
      if i2 <> i then x :: loop (Nat.succ i2) l else
      (l1 @ [x]) @ loop (Nat.succ i2) l in
  loop _0_ l2
;;

let rev_insert i l1 l2 =
  let rec loop accu i2 = function
    (* i2 is increasing from 0. *)
    | [] -> accu
    | x :: l ->
      let accu = if i2 <> i then x :: accu else x :: rev_append l1 accu in
      loop accu (Nat.succ i2) l in
  loop [] _0_ l2
;;

(* Iterators. *)

let rec iter_i_shift f i = function
  (* i is increasing. *)
  | [] -> ()
  | x :: l -> f i x; iter_i_shift f (Nat.succ i) l
;;

let iter_i f = iter_i_shift f _0_;;

let rec map_i_shift f i = function
  (* i is increasing. *)
  | [] -> []
  | x :: l -> let r = f i x in r :: map_i_shift f (Nat.succ i) l
;;

let map_i f = map_i_shift f _0_;;

let rec rev_map_i_end_shift accu f i = function
  (* i is increasing. *)
  | [] -> accu
  | x :: l -> rev_map_i_end_shift (f i x :: accu) f (Nat.succ i) l
;;

let rev_map_i f = rev_map_i_end_shift [] f _0_;;

let replace i0 x0 l = map_i (fun i x -> if i <> i0 then x else x0) l;;

let rev_replace i0 x0 l = rev_map_i (fun i x -> if i <> i0 then x else x0) l;;

let rev_flat_map_end accu f l =
  rev_append accu (fold_left (fun ll l -> rev_append (f l) ll) [] l)
;;

let rev_flat_map f = rev_flat_map_end [] f;;

let rec fold_left_i_shift f i accu = function
  (* i is increasing. *)
  | [] -> accu
  | x :: l ->
    let accu = f i accu x in
    fold_left_i_shift f (Nat.succ i) accu l
;;

let fold_left_i f = fold_left_i_shift f _0_;;

let rev_flat_map_i_end_shift accu f i l =
  rev_append
    accu
    (fold_left_i_shift (fun j ll l -> rev_append (f j l) ll) i [] l)
;;

let rev_flat_map_i f = rev_flat_map_i_end_shift [] f _0_;;

let flat_map_end accu f l = rev (rev_flat_map_end accu f l);;

let flat_map f = flat_map_end [] f;;

let flat_map_i_end_shift accu f i l =
  rev (rev_flat_map_i_end_shift accu f i l)
;;

let flat_map_i f = flat_map_i_end_shift [] f _0_;;

let concat ll = flat_map Upervasives.id ll;;

let flatten = concat;;

let rec fold_right_i_shift f i l accu =
  (* i is increasing. *)
  match l with
  | [] -> accu
  | x :: l ->
    let accu = f i x accu in
    fold_right_i_shift f (Nat.succ i) l accu
;;

let fold_right_i f = fold_right_i_shift f _0_;;

let rev_fold_left_env = fold_left;;

let fold_left_env f (env, accu) l =
  let (env, accu) = rev_fold_left_env f (env, rev accu) l in
  (env, rev accu)
;;

let fold_left_env_init f env l = fold_left_env f (env, []) l;;

(* Iterators on two lists. *)

let rec iter2_i_shift f i l1 l2 =
  (* i is increasing. *)
  match (l1, l2) with
  | ([], []) -> ()
  | (x1 :: l1, x2 :: l2) -> f i x1 x2; iter2_i_shift f (Nat.succ i) l1 l2
  | (_, _) -> invalid_arg "iter2_i_shift" "different lengths"
;;

let iter2_i f = iter2_i_shift f _0_;;

let rec map2_i_shift f i l1 l2 =
  (* i is increasing. *)
  match (l1, l2) with
  | ([], []) -> []
  | (x1 :: l1, x2 :: l2) ->
    let r = f i x1 x2 in
    r :: map2_i_shift f (Nat.succ i) l1 l2
  | (_, _) -> invalid_arg "map2_i_shift" "different lengths"
;;

let map2_i f = map2_i_shift f _0_;;

let rec rev_map2_i_end_shift accu f i l1 l2 =
  (* i is increasing. *)
  match (l1, l2) with
  | ([], []) -> accu
  | (x1 :: l1, x2 :: l2) ->
    rev_map2_i_end_shift (f i x1 x2 :: accu) f (Nat.succ i) l1 l2
  | (_, _) -> invalid_arg "rev_map2_i_end_shift" "different lengths"
;;

let rev_map2_i f = rev_map2_i_end_shift [] f _0_;;

let rev_flat_map2_end accu f l1 l2 =
  rev_append
    accu
    (fold_left2 (fun ll l1 l2 -> rev_append (f l1 l2) ll) [] l1 l2)
;;

let rev_flat_map2 f = rev_flat_map2_end [] f;;

let rec fold_left2_i_shift f i accu l1 l2 =
  (* i is increasing. *)
  match (l1, l2) with
  | ([], []) -> accu
  | (x1 :: l1, x2 :: l2) ->
    let accu = f i accu x1 x2 in
    fold_left2_i_shift f (Nat.succ i) accu l1 l2
  | (_, _) -> invalid_arg "fold_left2_i_shift" "different lengths"
;;

let fold_left2_i f = fold_left2_i_shift f _0_;;

let rev_flat_map2_i_end_shift accu f i l1 l2 =
  rev_append
    accu
    (fold_left2_i_shift
       (fun j ll l1 l2 -> rev_append (f j l1 l2) ll)
       i [] l1 l2)
;;

let rev_flat_map2_i f = rev_flat_map2_i_end_shift [] f _0_;;

let flat_map2_end accu f l1 l2 = rev (rev_flat_map2_end accu f l1 l2);;

let flat_map2 f = flat_map2_end [] f;;

let flat_map2_i_end_shift accu f i l1 l2 =
  rev (rev_flat_map2_i_end_shift accu f i l1 l2)
;;

let flat_map2_i f = flat_map2_i_end_shift [] f _0_;;

let rec fold_right2_i_shift f i l1 l2 accu =
  (* i is increasing. *)
  match (l1, l2) with
  | ([], []) -> accu
  | (x1 :: l1, x2 :: l2) ->
    let accu = f i x1 x2 accu in
    fold_right2_i_shift f (Nat.succ i) l1 l2 accu
  | (_, _) -> invalid_arg "fold_right2_i_shift" "different lengths"
;;

let fold_right2_i f = fold_right2_i_shift f _0_;;

let rev_fold_left2_env = fold_left2;;

let fold_left2_env f (env, accu) l1 l2 =
  let (env, accu) = rev_fold_left2_env f (env, rev accu) l1 l2 in
  (env, rev accu)
;;

let fold_left2_env_init f env l1 l2 = fold_left2_env f (env, []) l1 l2;;

(* Association lists. *)

let fold_assoc fun_name f = function
  | [] -> invalid_arg fun_name "empty list"
  | x0 :: l -> fold_left f x0 l
;;

let find_fst_assoc fun_name p =
  fold_assoc fun_name
    (fun (y0, _ as x0) (y, _ as x) -> if p y0 y then x else x0)
;;

let compare_assoc cmp =
  find_fst_assoc "compare_assoc" (fun y0 y -> cmp y0 y > 0)
;;

let minus_pervasives_compare x y = - Pervasives.compare x y;;

let min_assoc l = compare_assoc Pervasives.compare l
and max_assoc l = compare_assoc minus_pervasives_compare l
;;

let compare_fst_assoc = compare_assoc
and min_fst_assoc = min_assoc
and max_fst_assoc = max_assoc
;;

let find_snd_assoc fun_name p =
  fold_assoc fun_name
    (fun (_, y0 as x0) (_, y as x) -> if p y0 y then x else x0)
;;

let compare_snd_assoc cmp =
  find_snd_assoc "compare_snd_assoc" (fun y0 y -> cmp y0 y > 0)
;;

let min_snd_assoc l = compare_snd_assoc Pervasives.compare l
and max_snd_assoc l = compare_snd_assoc minus_pervasives_compare l
;;

(* Sorting. *)

let of_range pred l r =
  if r <= l then [] else
  (* Now, we have l < r.
     Therefore, the following pred will not fail. *)
  let r = pred r in
  let rec loop accu s =
    (* s is decreasing from r. *)
    if s < l then accu else
    if s = l then l :: accu else
    (* Now, we have l < s.
       Therefore, the following pred will not fail. *)
    loop (s :: accu) (pred s) in
  loop [] r
;;
