(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: umatrix.ml,v 1.50 2017/11/15 11:53:02 fclement Exp $ *)

(* Operations on matrices. *)

module Nat = Unatural;;
module L = Ulist;;
module A = Uarray;;
module Mt = Umatrix_types;;

let _0_ = Nat._0_;;

let nat2_zero = (_0_, _0_)
and nat2_add (a, b) (c, d) = (Nat.add a c, Nat.add b d)
and nat2_mul (a, b) (c, d) = (Nat.mul a c, Nat.mul b d)
and nat2_succ (a, b) = (Nat.succ a, Nat.succ b)
and nat2_transpose (a, b) = (b, a)
and nat2_pred (a, b) = (Nat.pred a, Nat.pred b)
and nat2_rails (fa, fb) (a, b) = (fa a, fb b)
;;

let nat_scale n1 n2 n = Nat.div (Nat.mul n n2) n1;;
let nat2_scale (a1, b1) (a2, b2) =
  nat2_rails (nat_scale a1 a2, nat_scale b1 b2)
;;

let curry f x y = f (x, y);;

let qual = Upervasives.qualify "Umatrix";;
let invalid_arg fun_name = Upervasives.invalid_arg_qual qual fun_name;;

type 'a t = 'a Mt.t
and 'a matrix = 'a t
;;

let of_array_array = Mt.to_t;;
external unsafe_of_array_array : 'a array array -> 'a t = "%identity";;
external to_array_array : 'a t -> 'a array array = "%identity";;

let oaa = of_array_array
and uoaa = unsafe_of_array_array
and taa = to_array_array
;;

type length = A.length
and index = A.index
and offset = A.offset

and line_number = length
and column_number = length
and line_index = index
and column_index = index
and cell_index = index
and line_offset = offset
and column_offset = offset
;;

type size = line_number * column_number
and index2 = line_index * column_index
and cell = index2
and offset2 = line_offset * column_offset
;;

let size = Mt.size
and dim_i = Mt.dim_i
and dim_j = Mt.dim_j
and set_line = Mt.set_line
and empty = Mt.empty
;;

let length m =
  let (ni, nj) = size m in
  Nat.mul ni nj
;;

let unsafe_get_line m = A.unsafe_get (taa m);;

let unsafe_get m i = A.unsafe_get (unsafe_get_line m i)
and unsafe_set m i = A.unsafe_set (unsafe_get_line m i)
;;

let unsafe_get_column m j =
  let ni = dim_i m in
  A.init ni (fun i -> unsafe_get m i j)
;;

let unsafe_set_column m j aj =
  A.iter_i (fun i x -> unsafe_set m i j x) aj
;;

let unsafe_get2 m (i, j) = unsafe_get m i j
and unsafe_set2 m (i, j) = unsafe_set m i j
;;

let check_bound = Mt.check_bound
and check_line_bound = Mt.check_line_bound
and check_array_length = Mt.check_array_length
;;

let check_column_bound = check_bound "column";;

let check_bounds2 fun_name (i, j) (ni, nj) =
  check_line_bound fun_name i ni;
  check_column_bound fun_name j nj
;;

let get2 m c =
  check_bounds2 "get" c (size m);
  (* Now, we have c = (i, j) with 0 <= i < ni and 0 <= j < nj.
     Therefore, the following unsafe_get2 will be safe. *)
  unsafe_get2 m c

and set2 m c =
  check_bounds2 "set" c (size m);
  (* Now, we have c = (i, j) with 0 <= i < ni and 0 <= j < nj.
     Therefore, the following unsafe_set2 will be safe. *)
  unsafe_set2 m c
;;

let get m i j = get2 m (i, j)
and set m i j = set2 m (i, j)
;;

let get_line m i =
  check_line_bound "get_line" i (dim_i m);
  (* Now, we have 0 <= i < dim_i m.
     Therefore, the following unsafe_get_line will be safe. *)
  unsafe_get_line m i
;;

let get_column m j =
  check_column_bound "get_column" j (dim_j m);
  (* Now, we have 0 <= j < dim_j m.
     Therefore, the following unsafe_get_column will be safe. *)
  unsafe_get_column m j
;;

let set_column m j aj =
  let (ni, nj) = size m in
  check_column_bound "set_column" j nj;
  check_array_length "set_column" aj ni;
  (* Now, we have 0 <= j < nj and length aj = ni.
     Therefore, the following unsafe_set_column will be safe. *)
  unsafe_set_column m j aj
;;

let make ni nj x = uoaa (A.init ni (fun _i -> A.make nj x))
and init ni nj f = uoaa (A.init ni (fun i -> A.init nj (f i)))
;;

let make2 (ni, nj) = make ni nj
and init2 (ni, nj) f = init ni nj (curry f)
;;

let check_dimension msg dim fun_name m1 m2 =
  if dim m1 <> dim m2 then
    invalid_arg fun_name (Printf.sprintf "%s dimension differ" msg)
;;

let check_first_dimension fun_name =
  check_dimension "first" dim_i fun_name
and check_second_dimension fun_name =
  check_dimension "second" dim_j fun_name
;;

let append_i m1 m2 =
  check_second_dimension "append_i" m1 m2;
  uoaa (A.append (taa m1) (taa m2))
;;

let append_j m1 m2 =
  check_first_dimension "append_j" m1 m2;
  uoaa (A.map2 A.append (taa m1) (taa m2))
;;

let append m1 m2 m3 m4 = append_i (append_j m1 m2) (append_j m3 m4);;

let list_check_dimension check_dim fun_name = function
  | [] -> ()
  | m :: ml -> L.iter (check_dim fun_name m) ml
;;

let list_check_first_dimension fun_name =
  list_check_dimension check_first_dimension fun_name
and list_check_second_dimension fun_name =
  list_check_dimension check_second_dimension fun_name
;;

let concat_i ml =
  list_check_second_dimension "concat_i" ml;
  oaa (A.concat (L.map taa ml))
;;

let concat_j ml =
  list_check_first_dimension "concat_j" ml;
  oaa (A.map_list A.concat (L.map taa ml))
;;

let concat mll = concat_i (L.map concat_j mll);;

let check_submatrix fun_name m (oi, oj) (li, lj) =
  let (ni, nj) = size m in
  if Nat.add oi li > ni || Nat.add oj lj > nj then
    invalid_arg fun_name "wrong submatrix specification"
;;

let sub2 m o2 sz =
  check_submatrix "sub" m o2 sz;
  (* Now, we have 0 <= oi <= oi + li <= ni and 0 <= oj <= oj + lj <= nj
     with o2 = (oi, oj) and sz = (li, lj).
     Therefore, the following unsafe_get2 will be safe for all c = (i, j)
     with 0 <= i < li and 0 <= j < lj. *)
  init2 sz (fun c -> unsafe_get2 m (nat2_add o2 c))
;;

let sub m oi li oj lj = sub2 m (oi, oj) (li, lj);;

(* In the following, we have 0 <= i < ni and 0 <= j < nj with c = (i, j) and
   size m = (ni, nj).
   Therefore, unsafe_get2 will be safe for all c. *)
let copy m = init2 (size m) (unsafe_get2 m);;

let for_loop2 (oi, oj) (li, lj) f =
  A.for_loop oi li (fun i -> A.for_loop oj lj (fun j -> f (i, j)))
;;

let gen_nat_for_loop2 fli flj (ifirst, jfirst) (ilast, jlast) f =
  fli ifirst ilast (fun i -> flj jfirst jlast (fun j -> f (i, j)))
;;

let nat_for_loop2 =
  gen_nat_for_loop2 A.nat_for_loop A.nat_for_loop

and nat_rev_for_loop_i2 =
  gen_nat_for_loop2 A.nat_rev_for_loop A.nat_for_loop

and nat_rev_for_loop_j2 =
  gen_nat_for_loop2 A.nat_for_loop A.nat_rev_for_loop

and nat_rev_for_loop_ij2 =
  gen_nat_for_loop2 A.nat_rev_for_loop A.nat_rev_for_loop
;;

let fill2 m o2 sz x =
  check_submatrix "fill" m o2 sz;
  (* Now, we have 0 <= oi <= oi + li < ni and 0 <= oj <= oj + lj < nj with
     o2 = (oi, oj), sz = (li, lj) and size m = (ni, nj).
     Therefore, the following unsafe_set2 will be safe for all c = (i, j)
     with oi <= i < oi + li and oj <= j < oj + lj. *)
  for_loop2 o2 sz (fun c -> unsafe_set2 m c x)
;;

let fill m oi li oj lj = fill2 m (oi, oj) (li, lj);;

let blit2 m1 (oi1, oj1 as o21) m2 (oi2, oj2 as o22) (li, lj as sz) =
  if li = _0_ || lj = _0_ then () else
  (* Now, we have li, lj > 0.
     Therefore, the following Nat.pred will not fail. *)
  check_submatrix "blit" m1 o21 sz;
  check_submatrix "blit" m2 o22 sz;
  (* Now, we have 0 <= oi1 <= oi1 + li < ni, 0 <= oi2 <= oi2 + li < ni,
     0 <= oj1 <= oj1 + lj < nj, and 0 <= oj2 <= oj2 + lj < nj.
     Therefore, the following unsafe_get2 and unsafe_set2 will be safe for
     all c = (i, j) with 0 <= i <= li - 1 and 0 <= j <= lj - 1. *)
  let bl c =
    let m1c = unsafe_get2 m1 (nat2_add o21 c) in
    unsafe_set2 m2 (nat2_add o22 c) m1c in
  let pli = Nat.pred li
  and plj = Nat.pred lj in
  let loop_ij =
    if oi1 < oi2
    then
      if oj1 < oj2
      then nat_rev_for_loop_ij2 (pli, _0_) (plj, _0_)
      else nat_rev_for_loop_i2 (pli, _0_) (_0_, plj)
    else
      if oj1 < oj2
      then nat_rev_for_loop_j2 (_0_, pli) (plj, _0_)
      else nat_for_loop2 (_0_, pli) (_0_, plj) in
  loop_ij bl
;;

let blit m1 oi1 oj1 m2 oi2 oj2 li lj =
  blit2 m1 (oi1, oj1) m2 (oi2, oj2) (li, lj)
;;

let is_empty m = dim_i m = _0_;;

let check_nonempty fun_name m =
  if is_empty m then Upervasives.invalid_arg fun_name "empty matrix"
;;

(* In the following, we have 0 <= i < ni and 0 <= j < nj with c = (i, j) and
   size m = (ni, nj).
   Therefore, unsafe_get2 will be safe for all c. *)
let transpose m =
  init2 (nat2_transpose (size m)) (fun c -> unsafe_get2 m (nat2_transpose c))
;;

(* Converters. *)

let cell_index nj (i, j) = Nat.add (Nat.mul nj i) j;;

let cell nj ii = (Nat.div ii nj, Nat.rem ii nj);;

let to_array m = A.concat (A.to_list (taa m));;

let of_array2 a (ni, nj as sz) =
  let len = A.length a in
  if len <> Nat.mul ni nj then
    invalid_arg "of_array" "wrong size specification" else
  (* FIXME: we need a proof here! *)
  init2 sz (fun c -> A.unsafe_get a (cell_index nj c))
;;

let of_array a ni nj = of_array2 a (ni, nj);;

let to_array_list m = A.to_list (taa m);;

let of_array_list al = oaa (A.of_list al);;

let to_list m = L.map A.to_list (A.to_list (taa m));;

let of_list l = oaa (A.map A.of_list (A.of_list l));;

(** {6 Iterators} *)

let iter f m = A.iter (A.iter f) (taa m);;

let map f m = oaa (A.map (A.map f) (taa m));;

let iter_ij f m =
  A.iter_i (fun i -> A.iter_i (fun j x -> f i j x)) (taa m)
;;

let map_ij f m =
  oaa (A.map_i (fun i -> A.map_i (fun j x -> f i j x)) (taa m))
;;

let iter_ij2 f = iter_ij (curry f)
and map_ij2 f = map_ij (curry f)
;;

let iter_i f m = A.iter f (taa m);;

let iter_j f m = iter_i f (transpose m);;

let map_i f m = oaa (A.map f (taa m));;

let map_j f m = transpose (map_i f (transpose m));;

let fold_left f accu m =
  A.fold_left (A.fold_left f) accu (taa m)
;;

let fold_right f m =
  A.fold_right (A.fold_right f) (taa m)
;;

let sub_iter f m oi li oj lj =
  A.sub_iter (fun mi -> A.sub_iter f mi oj lj) (taa m) oi li
;;

let sub_iter_ij f m oi li oj lj =
  A.sub_iter_i
    (fun i mi -> A.sub_iter_i (f i) mi oj lj)
    (taa m) oi li
;;

let sub_iter2 f m (oi, oj) (li, lj) = sub_iter f m oi li oj lj
and sub_iter_ij2 f m (oi, oj) (li, lj) = sub_iter_ij (curry f) m oi li oj lj
;;

(** {6 Iterators on two matrices} *)

let check_same_size fun_name m1 m2 =
  let sz = size m1 in
  if size m2 <> sz then
    invalid_arg fun_name "matrix arguments have different size" else
  sz
;;

let iter2 f m1 m2 =
  let _ = check_same_size "iter2" m1 m2 in
  A.iter2 (A.iter2 f) (taa m1) (taa m2)

and map2 f m1 m2 =
  let _ = check_same_size "map2" m1 m2 in
  oaa (A.map2 (A.map2 f) (taa m1) (taa m2))

and iter2_ij f m1 m2 =
  let _ = check_same_size "iter2_ij" m1 m2 in
  A.iter2_i (fun i -> A.iter2_i (f i)) (taa m1) (taa m2)

and map2_ij f m1 m2 =
  let _ = check_same_size "map2_ij" m1 m2 in
  oaa (A.map2_i (fun i -> A.map2_i (f i)) (taa m1) (taa m2))
;;

let iter2_ij2 f = iter2_ij (curry f)
and map2_ij2 f = map2_ij (curry f)
;;

let fold_left2 f accu m1 m2 =
  A.fold_left2 (A.fold_left2 f) accu (taa m1) (taa m2)
;;

let fold_right2 f m1 m2 =
  A.fold_right2 (A.fold_right2 f) (taa m1) (taa m2)
;;

(* Matrix scanning. *)

let for_all p m = A.for_all (A.for_all p) (taa m);;

let exists p m = A.exists (A.exists p) (taa m);;

let mem x = exists (( = ) x)
and memq x = exists (( == ) x)
;;

(* Selecting elements in matrices. *)

let map_select f chi m = A.map2 (A.map_select f) (taa chi) (taa m);;

let select chi m = A.map2 A.select (taa chi) (taa m);;

(* In the following, the resulting array array is necessarily of the same
 size as chi. Therefore, the following unsafe_of_array_array (uoaa) will not
 fail. *)
let unselect x chi aa = uoaa (A.map2 (A.unselect x) (taa chi) aa);;

(* Specific applications. *)

let make_diagonal_matrix zero v =
  let n = A.length v in
  (* In the following, we have 0 <= i < n = length v.
     Therefore, the following A.unsafe_get will be safe. *)
  init n n (fun i j -> if i = j then A.unsafe_get v i else zero)
;;

let make_int_diagonal_matrix v = make_diagonal_matrix 0 v;;
let make_float_diagonal_matrix v = make_diagonal_matrix 0. v;;

let resize2 (ni2, nj2 as sz2) m1 =
  let len1 = length m1 in
  let len2 = Nat.mul ni2 nj2 in
  if len2 <> len1 || (nj2 = _0_ && ni2 > _0_) then
    invalid_arg "resize" "wrong size specification" else
  let (ni1, nj1 as sz1) = size m1 in
  if ni1 = _0_ then empty else begin
    (* Now, we have ni1 > 0, thus m1 is not the empty matrix and we also have
       nj1 > 0.
       Therefore, the following unsafe_get2 will be safe. *)
    assert (nj1 > _0_);
    let m2 = make2 sz2 (unsafe_get2 m1 nat2_zero) in
    let fc = nat2_scale sz1 sz2 in
    iter_ij2
      (fun c1 x ->
        let c1first = fc c1
        and c1last = nat2_pred (fc (nat2_succ c1)) in
        (* FIXME: we need a proof here! *)
        nat_for_loop2 c1first c1last (fun c2 -> unsafe_set2 m2 c2 x))
      m1;
    m2
  end
;;

let zoom2 sz m1 = resize2 (nat2_mul sz (size m1)) m1;;

let resize ni = curry resize2 ni
and zoom ni = curry zoom2 ni
;;
