(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: unatural.ml,v 1.30 2017/11/15 12:02:16 fclement Exp $ *)

(* Natural numbers. *)

module P = Pervasives;;
module Nt = Unatural_types;;

let invalid_arg fun_name msg =
  let msg =
    if String.length msg = 0 then msg else
    Printf.sprintf ": %s" msg in
  Pervasives.invalid_arg (Printf.sprintf "Unatural.%s%s" fun_name msg)
;;

type t = Nt.t
and natural = t
;;

let of_int = Nt.to_t;;
external unsafe_of_int : int -> t = "%identity";;
external to_int : t -> int = "%identity";;

(* The following application of unsafe_of_int take nonnegative arguments.
   Therefore, they will not fail. *)
let _0_ = unsafe_of_int 0
and _1_ = unsafe_of_int 1
and _2_ = unsafe_of_int 2
and _3_ = unsafe_of_int 3
and _4_ = unsafe_of_int 4
and _5_ = unsafe_of_int 5
and _6_ = unsafe_of_int 6
and _7_ = unsafe_of_int 7
and _8_ = unsafe_of_int 8
and _9_ = unsafe_of_int 9
;;

let lift f x = of_int (f (to_int x))
and lift2 f x y = of_int (f (to_int x) (to_int y))
;;

let add = lift2 P.( + )
and sub = lift2 P.( - )
and mul = lift2 P.( * )
and div = lift2 P.( / )
and rem = lift2 P.( mod )
;;

let rec gcd p q =
  if q = _0_ then p else
  gcd q (rem p q)
;;

let succ = lift P.succ
and pred = lift P.pred
;;

(* The following application of unsafe_of_int takes a nonnegative argument.
   Therefore, it will not fail. *)
let max_nat = unsafe_of_int P.max_int
and min_nat = _0_
;;

let of_float f = of_int (P.int_of_float f)
and to_float n = P.float_of_int (to_int n)
;;

external of_char : char -> t = "%identity";;
external unsafe_to_char : t -> char = "%identity";;

let to_char n =
  if to_int n > 0xFF then invalid_arg "to_char" "" else
  (* Now, we have 0 <= n <= 0xFF = 255.
     Therefore, the following unsafe_to_char will not fail. *)
  unsafe_to_char n
;;

let of_string s = of_int (P.int_of_string s)
and to_string s = P.string_of_int (to_int s)
;;

let compare x y = P.( - ) (to_int x) (to_int y);;

let ( // ) a b =
  let q = div a b in
  if a = mul b q then q else
  succ q
;;

let split n k =
  let rec loop accu s nn kk =
    (* s is increasing from 0, nn is decreasing from n, and kk is decreasing
       from k. *)
    if s = k then accu else
    if nn = _0_ then loop (_0_ :: accu) (succ s) nn kk else
    (* Now, we have nn > 0.
       Moreover, kk decreases by one at each loop iteration, and as soon as
       kk = 1, then q = nn // 1 = nn, and we loop with nn = 0.
       Thus, we also have kk > 0.
       Therefore, the following pred will not fail. *)
    let q = nn // kk in
    loop (q :: accu) (succ s) (sub nn q) (pred kk) in
  loop [] _0_ n k
;;

let is_zero x = x = _0_
and is_positive x = x <> _0_
and is_even x = rem x _2_ = _0_
and is_odd x = rem x _2_ = _1_
;;

let psplit n k = List.filter is_positive (split n k);;

let incr r = r := succ !r
and decr r = r := pred !r
;;

type tmin = t
and tmax = t
;;

let for_loop ifirst ilast body =
  for i = to_int ifirst to to_int ilast do
    (* Now, we have 0 <= ifirst <= i <= ilast <= max_nat = max_int.
       Therefore, the following unsafe_of_int will not fail. *)
    body (unsafe_of_int i);
  done;
;;

let rev_for_loop ifirst ilast body =
  for i = to_int ifirst downto to_int ilast do
    (* Now, we have 0 <= ifirst <= i <= ilast <= max_nat = max_int.
       Therefore, the following unsafe_of_int will not fail. *)
    body (unsafe_of_int i);
  done;
;;
