(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: upervasives.mli,v 1.45 2017/10/27 10:43:36 fclement Exp $ *)

(** {3 Extensions to the initially opened module} *)

(** {6 Byte sequences} *)

type bytes = Upervasives_types.bytes;;

(** {6 Functional helpers} *)

external id : 'a -> 'a = "%identity"
(** [id x] returns [x], it is the identity function. *)
;;

val compose : ('b -> 'c) -> ('a -> 'b) -> 'a -> 'c;;

val proj2_1 : 'a -> 'b -> 'a;;
val proj2_2 : 'a -> 'b -> 'b;;

val exercise_option : 'a -> 'a option -> 'a
(** [exercise_option default opt] either returns [x] when [opt] is [Some x],
 or [default] otherwise. *)
;;

val liftl_option : 'b -> ('a -> 'b) -> 'a option -> 'b
(** [liftl_option default f opt] either returns [f x] when [opt] is [Some x],
 or returns [default] otherwise. *)
;;

val liftr_option : ('a -> 'b) -> 'a -> 'b option
(** [liftr_option f x] returns [Some (f x)]. *)
;;

val downl_option : ('a option -> 'b) -> 'a -> 'b
(** [downl_option f x] returns [f (Some x)]. *)
;;

val downr_option : 'b -> ('a -> 'b option) -> 'a -> 'b
(** [downr_option default f x] either returns [y] when [f x] is [Some y], or
 returns [default] otherwise. *)
;;

val lift_option : ('a -> 'b) -> 'a option -> 'b option
(** [lift_option f opt] either returns [Some (f x)] when [opt] is [Some x],
 or returns [None] otherwise. *)
;;

val down_option : 'b -> ('a option -> 'b option) -> 'a -> 'b
(** [down_option default f x] either returns [y] when [f (Some x)] is [Some y],
 or [default] otherwise. *)
;;

(** {6 Generating messages} *)

type software_name = string
and module_name = string
and function_name = string
and tag = string
and prefix = string
and message = string

and gen_message = unit -> message
and qualify = function_name -> function_name
;;

val prefix : prefix -> message -> message;;
val qualify : module_name -> qualify;;
val quote : function_name -> message -> message;;

val report_message : tag option -> message -> message;;
val debug_message : tag option -> function_name option -> message -> message;;
val warning_message : tag option -> function_name option -> message -> message;;
val error_message : tag option -> function_name option -> message -> message
(** Functions generating messages of the following form:
 * report:  "<tag> <message>"
 * debug:   "<tag> <function name>: <message>"
 * warning: "<tag> Warning in <function name>: <message>"
 * error:   "<tag> Fatal error in <function name>: <message>" *)
;;

(** {6 Raising exceptions} *)

val invalid_arg : function_name -> message -> 'a;;
val failwith : function_name -> message -> 'a
(** Raise [Invalid_arg] or [Failure] with the message computed by
 [error_message] applied to the same arguments. *)
;;

val internal_error : function_name -> 'a
(** Raise [Failure] with the message computed by [error_message] applied
 to the same arguments and ["Internal error, please report."]. *)
;;

val invalid_arg_qual : qualify -> function_name -> message -> 'a;;
val failwith_qual : qualify -> function_name -> message -> 'a;;
val internal_error_qual : qualify -> function_name -> 'a
(** Same as [invalid_arg], [failwith], and [internal_error],
 but with a qualification function as an additional first argument,
 e.g. a partial application of {!qualify} to some module name. *)
;;

(** {6 Integer arithmetic} *)

type length = Unatural.t;;

val length_of_int : int -> length;;
external length_to_int : length -> int = "%identity";;

type index = Unatural.t;;

val index_of_int : int -> index;;
external index_to_int : index -> int = "%identity";;

type offset = Unatural.t;;

val offset_of_int : int -> offset;;
external offset_to_int : offset -> int = "%identity";;

(** {6 Floating-point arithmetic} *)

val round : float -> int
(** Round the given float to the nearest integer.
  For a nonnegative float [f], [round f] returns Pervasives.int_of_float
  [(f +. 0.5)], ie the greatest integer less than or equal to [f +. 0.5];
  and for a negative float [f], [round f] returns Pervasives.int_of_float
  [(f -. 0.5)], ie the least integer greater than or equal to [f -. 0.5]. *)
;;

(** {6 Conversion} *)

val bit_string_of_float : float -> string
(** [bit_string_of_float f] returns a [string] containing the string of bits
  of the binary representation of the float [f]. *)
;;

val float_of_bit_string : string -> float
(** [float_of_bit_string] is the converse function of [bit_string_of_float]. *)
;;

(** {6 Data selection} *)

type select =
  | Keep
  | Forget
(** Intention to select data structure elements. *)
;;

(** {6 Input/output} *)

val in_channel_length : in_channel -> length
(** Return the length (number of characters) of the regular file on which the
 given channel is opened.
 If the channel is opened on a file that is not a regular file, the result is
 meaningless.
 The returned length does not take into account the end-of-line translations
 that can be performed when reading from a channel opened in text mode. *)
;;

val input_string : length -> in_channel -> string
(** [input_string len ic] reads characters from input channel [ic].
 The string result contains [len] characters or all the available characters
 if there are not enough available characters in [ic].

 Raise [End_of_file] if there are no characters to read. *)
;;

val string_of_ic : in_channel -> string
(** [string_of_ic ic] reads all the available characters in input channel [ic].
 Returns the empty string if there are no characters to read. *)
;;

val iter_channel : (string -> unit) -> in_channel -> unit
(** [iter_channel ic f] iterates function [f] on all successive string chunks
 read in input channel [ic]. *)
;;

val print_unit : unit -> unit
(** Print unit on standard output. *)
;;

val prerr_unit : unit -> unit
(** Print unit on standard error. *)
;;

val print_bool : bool -> unit
(** Print a boolean on standard output. *)
;;

val prerr_bool : bool -> unit
(** Print a boolean on standard error. *)
;;

val open_append : string -> out_channel;;
val open_append_bin : string -> out_channel;;

(**/**)
(** {6 Undocumented features} *)

type gen_message_option = unit -> message option
and qualify_opt = function_name option -> function_name option
;;

val prefix_opt : prefix option -> message option -> message option;;
val qualify_opt : module_name option -> qualify_opt;;
val quote_opt : function_name option -> message option -> message option;;
val report_message_opt :
  tag option -> gen_message_option -> message option
;;
val debug_message_opt :
  tag option -> function_name option -> gen_message_option -> message option
;;
val warning_message_opt :
  tag option -> function_name option -> gen_message_option -> message option
;;
val error_message_opt :
  tag option -> function_name option -> gen_message_option -> message option
(** Generic functions where arguments and results are options. *)
;;

val invalid_arg_opt :
  tag option -> function_name option -> gen_message_option -> 'a
;;
val failwith_opt :
  tag option -> function_name option -> gen_message_option -> 'a
;;
val internal_error_opt :
  tag option -> function_name option -> 'a
;;

val invalid_arg_qual_opt :
  tag option -> qualify_opt -> function_name option -> gen_message_option -> 'a
;;
val failwith_qual_opt :
  tag option -> qualify_opt -> function_name option -> gen_message_option -> 'a
;;
val internal_error_qual_opt :
  tag option -> qualify_opt -> function_name option -> 'a
(** Generic functions where arguments and results are options. *)
;;

external unsafe_length_of_int : int -> length = "%identity";;
external unsafe_index_of_int : int -> index = "%identity";;
external unsafe_offset_of_int : int -> offset = "%identity";;
