(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: upph_id_card.mli,v 1.34 2017/05/31 15:24:06 fclement Exp $ *)

(** {3 Handling the user part of the id card of a Pph project}

 This module is implemented over module {!Uoption_key}, {!Ucommand_option},
 and {!Uversion} (from the OCaml user's additional library Ulib). *)

type t = {
  project_name : string;
  project_banner_title : string;
  project_outline : string;
  project_release_date : string;
  project_copyright_holders : string;
  project_authors : string;
  project_http_url : string;
  project_ftp_url : string;
  project_release_ident : string;
  project_distribution_ident : string;
  project_version : string;
  project_version_full_info : string;
  project_release_name : string;
  project_distribution_name : string;
}
and pph_id_card = t
(** The type for Pph id cards.
 The intent is to automatically fill such a record from the information
 provided in file "config/project.env" for Pph projects, e.g. by using
 the generic text file generator Htmlc, also provided for Pph projects. *)
;;

(** {6 Printing}

 Print some version information provided by the [pph_id_card] argument on
 [stderr]. *)

val print_long_version_info : t -> unit
(** Print contents of fields [project_name], [project_banner_title] and
 [project_version_full_info] with separator [", "]. *)
;;

val print_version_info : t -> unit
(** Print contents of fields [project_name] and [project_release_ident] with
 separator [", "]. *)
;;

val print_short_version_info : t -> unit
(** Print contents of field [project_version]. *)
;;

(** {6 Command options} *)

val add_app_version_options : t -> Uoption_key.option_keys option -> unit
(** Add command options "<optional prefix>-long-version-info",
 "<optional prefix>-version-info" and "<optional prefix>-short-version" that
 apply the printing funtions above and exit.
 The optional prefix (starting with a minus character ['-']) is computed from
 the [Uoption_key.option_keys option] argument;
 the intent is to reflect somehow the name of the Pph software.
 Use this function when the Pph software is an application. *)
;;

val add_lib_version_options : t -> Uoption_key.option_keys option -> unit
(** Same as [add_app_version_options] but for libraries. *)
;;

val make_option_keys :
  t -> Uoption_key.short_option_key_status -> Uoption_key.option_keys
(** [make_option_keys pic] is a shortcut for {!Uoption_key.make_option_keys}
  [pic.project_name]. *)
;;

val add_app_version_options_with_sk :
  t -> Uoption_key.short_option_key_status -> unit
;;
val add_lib_version_options_with_sk :
  t -> Uoption_key.short_option_key_status -> unit
(** Same as above, but where the [long_option_key] prefix is automatically
 obtained using [make_option_keys]. *)
;;

(** {6 Managing automatic service options} *)

val long_version_info_app_clear : unit -> unit;;
val long_version_info_lib_clear : unit -> unit;;
val version_info_app_clear : unit -> unit;;
val version_info_lib_clear : unit -> unit;;
val short_version_info_app_clear : unit -> unit;;
val short_version_info_lib_clear : unit -> unit
(** Clear already added entries in the corresponding class of command
 options (see module {!Ucommand_option}). *)
;;

val long_version_info_app_clear_option_key : Uoption_key.t -> unit;;
val long_version_info_lib_clear_option_key : Uoption_key.t -> unit;;
val version_info_app_clear_option_key : Uoption_key.t -> unit;;
val version_info_lib_clear_option_key : Uoption_key.t -> unit;;
val short_version_info_app_clear_option_key : Uoption_key.t -> unit;;
val short_version_info_lib_clear_option_key : Uoption_key.t -> unit
(** Clear all already added entries in the corresponding class of command
 options which main option name matches the given [key] (see module
 {!Ucommand_option}). *)
;;

val phase_ulib_versioning : Ucommand_option.phase_name
(** Ulib versioning phase name. *)
;;

val phase_long_version_info : unit -> unit;;
val phase_version_info : unit -> unit;;
val phase_short_version_info : unit -> unit
(** Add all declared command options of the underlying class to the phase
 {!phase_ulib_versioning}, including automatically added service command
 option. Should only be used when all command options are declared, just
 before parsing the command line. *)
;;

val phase_all_versioning : unit -> unit
(** Shortcut to call all three [phase_*] functions above. *)
;;

(** {6 Shortcuts} *)

val application_printing_facilities :
  t -> Uoption_key.option_keys option -> Udebug.printing_facilities
;;
val library_printing_facilities :
  t -> Uoption_key.option_keys option -> Udebug.printing_facilities
;;
val application_printing_facilities_with_static_tag :
  t -> Uoption_key.option_keys option -> Udebug.printing_facilities
;;
val library_printing_facilities_with_static_tag :
  t -> Uoption_key.option_keys option -> Udebug.printing_facilities
;;
val application_printing_facilities_with_dynamic_tags :
  t -> Uoption_key.option_keys option ->
  'a Udebug.make_tag_opt -> 'a Udebug.dynamic_printing_facilities
;;
val library_printing_facilities_with_dynamic_tags :
  t -> Uoption_key.option_keys option ->
  'a Udebug.make_tag_opt -> 'a Udebug.dynamic_printing_facilities
(** Same as corresponding functions from {!Udebug} module, but where argument
 of type Ucommand_option.software_option_key_data is replaced by a first
 argument of type [pph_id_card] and a second of type
 {!Uoption_key.option_keys} [option]. *)
;;
