(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ustamped_darray.ml,v 1.15 2017/05/31 15:24:06 fclement Exp $ *)

(* Stamped dynamic arrays. *)

module Nat = Unatural;;
module A = Uarray;;
module Da = Udynamic_array;;

let _0_ = Nat._0_;;

let qual = Upervasives.qualify "Ustamped_darray";;
let invalid_arg = Upervasives.invalid_arg_qual qual;;

type index = Da.index;;
type length = Da.length;;

type 'a stamped =
  | Unset
  | Valid of 'a
;;

exception Not_set;;

type 'a t =
  {
   darray : 'a stamped Da.t;
   mutable next_unset : index;
  }
and 'a stamped_darray = 'a t
;;

let make n =
  {
   darray = Da.make n Unset;
   next_unset = _0_;
  }
;;

let get sda i =
  match Da.get sda.darray i with
  | Unset -> raise Not_set
  | Valid x -> x
;;

let set ({ darray = da; next_unset = n; } as sda) i x =
  if i > n then
    invalid_arg "set"
      (Printf.sprintf "index %d is greater then next unset %d"
         (Nat.to_int i) (Nat.to_int n)) else
  begin Da.set da i (Valid x);
  if i = n then sda.next_unset <- Nat.succ i end
;;

let set_next sda x = set sda sda.next_unset x;;

let same sda1 sda2 =
  Da.same sda1.darray sda2.darray &&
  sda1.next_unset = sda2.next_unset
;;

let equal_items equal_item sda1 sda2 =
  let equal_item st1 st2 =
    match (st1, st2) with
    | (Unset, Unset) -> true
    | (Unset, Valid _) | (Valid _, Unset) -> false
    | (Valid x1, Valid x2) -> equal_item x1 x2 in
  Da.equal_items equal_item sda1.darray sda2.darray
;;

let max_length = Da.max_length;;

let length sda = Da.length sda.darray;;

(* Converters. *)

let get_valid = function
  | Unset -> assert false
  | Valid x -> x
;;

let to_array { darray = da; next_unset = n; } =
  A.init n (fun i -> get_valid (Da.get da i))
;;

let valid x = Valid x;;

let of_array a =
  let n = A.length a in
  { darray = Da.of_array Unset (A.map valid a); next_unset = n; }
;;

(* Iterators. *)

let iter f { darray = da; _ } =
  let f i = function
    | Unset -> ()
    | Valid x -> f i x in
  Da.iter f da
;;

let map f ({ darray = da; _ } as sda) =
  let f i = function
    | Unset -> Unset
    | Valid x -> Valid (f i x) in
  { sda with darray = Da.map f da; }
;;

let fold_left f accu { darray = da; _ } =
  let f accu i = function
    | Unset -> accu
    | Valid x -> f accu i x in
  Da.fold_left f accu da
;;
