(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uuniverse.ml,v 1.22 2017/10/25 13:01:01 fclement Exp $*)

module Nat = Unatural;;
module L = Ulist;;
module L1 = Ulist1;;
module A = Uarray;;
module Idx = Uindex;;
module Ut = Uuniverse_types;;

let _0_ = Nat._0_
and _1_ = Nat._1_
;;

let qual = Upervasives.qualify "Uuniverse";;
let invalid_arg = Upervasives.invalid_arg_qual qual;;

type 'a index = 'a Idx.t;;

type 'a indices = 'a Ut.indices;;
type 'a elements = 'a array;;

type 'a t = 'a Ut.t = private
  {
   indices : 'a indices;
   mutable elements : 'a elements;
   mutable cardinal : Nat.t;
  }
and 'a universe = 'a t
;;

(* Expansion phase, or transient state. *)

let make = Ut.make
and unsafe_make = Ut.unsafe_make
and copy = Ut.copy
and mem = Ut.mem
and index = Ut.index
and unsafe_add = Ut.unsafe_add
and stats = Ut.stats
;;

let singleton x =
  (* 1 > 0.
     Therefore, the following unsafe_make will be safe. *)
  unsafe_make _1_ x
;;

let add u x =
  if not (mem u x) then
    (* Now, we have x not in u.
       Therefore, the following unsafe_add will be safe. *)
    unsafe_add u x
;;

let product
    { elements = elts1; cardinal = n1; _; }
    { elements = elts2; cardinal = n2; _; } =
  (* Universes are nonempty, thus we have n1, n2, n1*n2 > 0.
     Therefore, in the following, A.unsafe_get and unsafe_make will be safe
     and Nat.pred will not fail. *)
  let x10 = A.unsafe_get elts1 _0_
  and x20 = A.unsafe_get elts2 _0_ in
  let u = unsafe_make (Nat.mul n1 n2) (x10, x20) in
  let len1 = Nat.pred n1 in
  let len2 = Nat.pred n2 in
  let f x1 x2 = unsafe_add u (x1, x2) in
  (* In the following, the resulting universe is obtained by iterating on all
     pairs of elements from both input universes, and these elements are
     accessed to via their indices i1 and i2 (both from 0 to the length of
     the universe minus 1).
     The case (0, 0) is already treated when creating the product universe
     using unsafe_make above.
     The first A.sub_iter treats cases (0, i2) with i2 = 1..(len2 - 1).
     The second A.sub_iter treats cases (i1, i2) with i1 = 1..(len1 - 1) and
     i2 = 0..(len - 1).
     Thus, all pairs are treated once and only once.
     Therefore, the unsafe_add in f will always be safe. *)
  A.sub_iter (f x10) elts2 _1_ len2;
  A.sub_iter (fun x1 -> A.iter (f x1) elts2) elts1 _1_ len1;
  u
;;

let of_list1 l1 =
  (* Now, we have L1.length l1 > 0.
     Therefore, the following unsafe_make will be safe. *)
  let u = unsafe_make (L1.length l1) (L1.hd l1) in
  (* Note that the input list may contain doubles.
     Therefore, we cannot use unsafe_add here. *)
  L.iter (add u) (L1.tl l1);
  u
;;

let of_array a =
  let len = A.length a in
  if len = _0_ then invalid_arg "of_array" "empty array" else
  (* Now, we have len > 0.
     Therefore, in the following, A.unsafe_get and unsafe_make will be safe
     and Nat.pred will not fail. *)
  let u = unsafe_make len (A.unsafe_get a _0_) in
  (* Note that the input array may contain doubles.
     Therefore, we cannot use unsafe_add here. *)
  A.sub_iter (add u) a _1_ (Nat.pred len);
  u
;;

let gen_range succ i j =
  let rec loop accu i =
    if i >= j then Ulist.rev accu else
    loop (i :: accu) (succ i) in
  loop [] i
;;

let gen_of_range succ i j =
  if i >= j then invalid_arg "of_range" "empty range" else
  of_list1 (i, (gen_range succ) (succ i) j)
;;

let of_range = gen_of_range succ;;

let of_nat_range = gen_of_range Nat.succ;;

let of_rectangle h w =
  let hu = of_range 0 h
  and wu = of_range 0 w in
  product hu wu
;;

let of_nat_rectangle h w =
  let hu = of_nat_range _0_ h
  and wu = of_nat_range _0_ w in
  product hu wu
;;

let indices u = gen_of_range Idx.succ Idx._0_ (Idx.of_nat u.cardinal);;

(* Steady state. *)

let cardinal u = u.cardinal;;

let valid_index (u : 'a t) (index : 'a index) =
  Idx.to_nat index < cardinal u
;;

let check_index fun_name u index =
  if not (valid_index u index) then
    Upervasives.invalid_arg
      fun_name
      (Printf.sprintf "index %d out of universe of %d element(s)"
         (Idx.to_int index)
         (Nat.to_int (cardinal u)))
;;

let element { elements = elts; _; } index = A.get elts (Idx.to_nat index);;

let elements { elements = elts; cardinal = n; _; } = A.sub elts _0_ n;;

let sub_elements u sub_indices =
  A.map (fun index -> element u index) sub_indices
;;

let fold_left f accu { elements = elts; cardinal = n; _; } =
  A.sub_fold_left f accu elts _0_ n
;;

let fold_left_i f accu { elements = elts; cardinal = n; _; } =
  A.sub_fold_left_i (fun accu i -> f accu (Idx.of_nat i)) accu elts _0_ n
;;

let iter f { elements = elts; cardinal = n; _; } = A.sub_iter f elts _0_ n;;

let iter_i f { elements = elts; cardinal = n; _; } =
  A.sub_iter_i (fun i x -> f (Idx.of_nat i) x) elts _0_ n
;;

let map_elements f { elements = elts; cardinal = n; _; } =
  (* Universes are nonempty, thus we have n > 0.
     Therefore, in the following, A.unsafe_get will be safe and Nat.pred will
     not fail. *)
  let res = A.make n (f (A.unsafe_get elts _0_)) in
  A.sub_iter_i
    (fun i x -> A.set res i (f x))
    elts _1_ (Nat.pred n);
  res
;;

let map f u = of_array (map_elements f u);;

let find p { elements = elts; cardinal = n; _; } = A.sub_find p elts _0_ n;;

let find_all p { elements = elts; cardinal = n; _; } =
  A.rev_sub_find_all p elts _0_ n
;;

let to_list u = A.to_list (elements u);;
